"""Quadrature schemes on cells

This module generates quadrature schemes on reference cells that integrate
exactly a polynomial of a given degree using a specified scheme.

Scheme options are:

  scheme="default"

  scheme="canonical" (collapsed Gauss scheme)

Background on the schemes:

  Keast rules for tetrahedra:
    Keast, P. Moderate-degree tetrahedral quadrature formulas, Computer
    Methods in Applied Mechanics and Engineering 55(3):339-348, 1986.
    http://dx.doi.org/10.1016/0045-7825(86)90059-9

  Xiao-Gimbutas rules for simplices:
    Xiao, H., and Gimbutas, Z. A numerical algorithm for the construction of
    efficient quadrature rules in two and higher dimensions, Computers &
    mathematics with applications 59(2): 663-676, 2010.
"""

# Copyright (C) 2011 Garth N. Wells
# Copyright (C) 2016 Miklos Homolya
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# First added:  2011-04-19
# Last changed: 2011-04-19

import numpy

from FIAT.quadrature import (QuadratureRule, FacetQuadratureRule, make_quadrature,
                             make_tensor_product_quadrature, map_quadrature)
from FIAT.reference_element import (HEXAHEDRON, QUADRILATERAL, TENSORPRODUCT,
                                    TETRAHEDRON, TRIANGLE, UFCTetrahedron,
                                    UFCTriangle, symmetric_simplex)
from FIAT.macro import MacroQuadratureRule


def create_quadrature(ref_el, degree, scheme="default", entity=None):
    """
    Generate quadrature rule for given reference element that will integrate a
    polynomial of order 'degree' exactly.

    For low-degree polynomials on triangles (<=50) and tetrahedra (<=15), this uses
    hard-coded rules, otherwise it falls back to a collapsed Gauss scheme on
    simplices.  On tensor-product cells, it is a tensor-product quadrature rule
    of the subcells.

    :arg ref_el: The FIAT cell to create the quadrature for.
    :arg degree: The degree of polynomial that the rule should integrate exactly.
    :kwarg scheme: The quadrature scheme, can be choosen from ["default", "canonical", "KMV"]
        "default" -> hard-coded scheme for low degree and collapsed Gauss scheme for high degree,
        "canonical" -> collapsed Gauss scheme,
        "KMV" -> spectral lumped scheme for low degree (<=5 on triangles, <=3 on tetrahedra).
    :kwarg entity: A tuple of entity dimension and entity id specifying the
        integration domain. If not provided, the domain is the entire cell.
    """
    if entity is not None:
        dimension, entity_id = entity
        sub_el = ref_el.construct_subelement(dimension)
        Q_ref = create_quadrature(sub_el, degree, scheme=scheme)
        return FacetQuadratureRule(ref_el, dimension, entity_id, Q_ref)

    if ref_el.is_macrocell():
        dimension = ref_el.get_dimension()
        sub_el = ref_el.construct_subelement(dimension)
        Q_ref = create_quadrature(sub_el, degree, scheme=scheme)
        return MacroQuadratureRule(ref_el, Q_ref)

    if ref_el.get_shape() == TENSORPRODUCT:
        try:
            degree = tuple(degree)
        except TypeError:
            degree = (degree,) * len(ref_el.cells)

        assert len(ref_el.cells) == len(degree)
        quad_rules = [create_quadrature(c, d, scheme)
                      for c, d in zip(ref_el.cells, degree)]
        return make_tensor_product_quadrature(*quad_rules)

    if ref_el.get_shape() in [QUADRILATERAL, HEXAHEDRON]:
        return create_quadrature(ref_el.product, degree, scheme)

    if degree < 0:
        raise ValueError("Need positive degree, not %d" % degree)

    if scheme == "default":
        if ref_el.get_shape() == TRIANGLE:
            return _triangle_scheme(ref_el, degree)
        elif ref_el.get_shape() == TETRAHEDRON:
            return _tetrahedron_scheme(ref_el, degree)
        else:
            return _fiat_scheme(ref_el, degree)
    elif scheme == "canonical":
        return _fiat_scheme(ref_el, degree)
    elif scheme == "KMV":  # Kong-Mulder-Veldhuizen scheme
        return _kmv_lump_scheme(ref_el, degree)
    else:
        raise ValueError("Unknown quadrature scheme: %s." % scheme)


def _fiat_scheme(ref_el, degree):
    """Get quadrature scheme from FIAT interface"""

    # Number of points per axis for exact integration
    num_points_per_axis = (degree + 1 + 1) // 2

    # Create and return FIAT quadrature rule
    return make_quadrature(ref_el, num_points_per_axis)


def _kmv_lump_scheme(ref_el, degree):
    """Specialized quadrature schemes for P < 6 for KMV simplical elements."""

    sd = ref_el.get_spatial_dimension()
    # set the unit element
    if sd == 2:
        T = UFCTriangle()
    elif sd == 3:
        T = UFCTetrahedron()
    else:
        raise ValueError("Dimension not supported")

    if degree == 1:
        x = ref_el.vertices
        w = numpy.arange(sd + 1, dtype=numpy.float64)
        if sd == 2:
            w[:] = 1.0 / 6.0
        elif sd == 3:
            w[:] = 1.0 / 24.0
        else:
            raise ValueError("Dimension not supported")
    elif degree == 2:
        if sd == 2:
            x = list(ref_el.vertices)
            for e in range(3):
                x.extend(ref_el.make_points(1, e, 2))  # edge midpoints
            x.extend(ref_el.make_points(2, 0, 3))  # barycenter
            w = numpy.arange(7, dtype=numpy.float64)
            w[0:3] = 1.0 / 40.0
            w[3:6] = 1.0 / 15.0
            w[6] = 9.0 / 40.0
        elif sd == 3:
            x = list(ref_el.vertices)
            x.extend(
                [
                    (0.0, 0.50, 0.50),
                    (0.50, 0.0, 0.50),
                    (0.50, 0.50, 0.0),
                    (0.0, 0.0, 0.50),
                    (0.0, 0.50, 0.0),
                    (0.50, 0.0, 0.0),
                ]
            )
            # in facets
            x.extend(
                [
                    (0.33333333333333337, 0.3333333333333333, 0.3333333333333333),
                    (0.0, 0.3333333333333333, 0.3333333333333333),
                    (0.3333333333333333, 0.0, 0.3333333333333333),
                    (0.3333333333333333, 0.3333333333333333, 0.0),
                ]
            )
            # in the cell
            x.extend([(1 / 4, 1 / 4, 1 / 4)])
            w = numpy.arange(15, dtype=numpy.float64)
            w[0:4] = 17.0 / 5040.0
            w[4:10] = 2.0 / 315.0
            w[10:14] = 9.0 / 560.0
            w[14] = 16.0 / 315.0
        else:
            raise ValueError("Dimension not supported")

    elif degree == 3:
        if sd == 2:
            alpha = 0.2934695559090401
            beta = 0.2073451756635909
            x = list(ref_el.vertices)
            x.extend(
                [
                    (1 - alpha, alpha),
                    (alpha, 1 - alpha),
                    (0.0, 1 - alpha),
                    (0.0, alpha),
                    (alpha, 0.0),
                    (1 - alpha, 0.0),
                ]  # edge points
            )
            x.extend(
                [(beta, beta), (1 - 2 * beta, beta), (beta, 1 - 2 * beta)]
            )  # points in center of cell
            w = numpy.arange(12, dtype=numpy.float64)
            w[0:3] = 0.007436456512410291
            w[3:9] = 0.02442084061702551
            w[9:12] = 0.1103885289202054
        elif sd == 3:
            x = list(ref_el.vertices)
            x.extend(
                [
                    (0, 0.685789657581967, 0.314210342418033),
                    (0, 0.314210342418033, 0.685789657581967),
                    (0.314210342418033, 0, 0.685789657581967),
                    (0.685789657581967, 0, 0.314210342418033),
                    (0.685789657581967, 0.314210342418033, 0.0),
                    (0.314210342418033, 0.685789657581967, 0.0),
                    (0, 0, 0.685789657581967),
                    (0, 0, 0.314210342418033),
                    (0, 0.314210342418033, 0.0),
                    (0, 0.685789657581967, 0.0),
                    (0.314210342418033, 0, 0.0),
                    (0.685789657581967, 0, 0.0),
                ]
            )  # 12 points on edges of facets (0-->1-->2)

            x.extend(
                [
                    (0.21548220313557542, 0.5690355937288492, 0.21548220313557542),
                    (0.21548220313557542, 0.21548220313557542, 0.5690355937288492),
                    (0.5690355937288492, 0.21548220313557542, 0.21548220313557542),
                    (0.0, 0.5690355937288492, 0.21548220313557542),
                    (0.0, 0.21548220313557542, 0.5690355937288492),
                    (0.0, 0.21548220313557542, 0.21548220313557542),
                    (0.5690355937288492, 0.0, 0.21548220313557542),
                    (0.21548220313557542, 0.0, 0.5690355937288492),
                    (0.21548220313557542, 0.0, 0.21548220313557542),
                    (0.5690355937288492, 0.21548220313557542, 0.0),
                    (0.21548220313557542, 0.5690355937288492, 0.0),
                    (0.21548220313557542, 0.21548220313557542, 0.0),
                ]
            )  # 12 points (3 points on each facet, 1st two parallel to edge 0)
            alpha = 1 / 6
            x.extend(
                [
                    (alpha, alpha, 0.5),
                    (0.5, alpha, alpha),
                    (alpha, 0.5, alpha),
                    (alpha, alpha, alpha),
                ]
            )  # 4 points inside the cell
            w = numpy.arange(32, dtype=numpy.float64)
            w[0:4] = 0.00068688236002531922325120561367839
            w[4:16] = 0.0015107814913526136472998739890272
            w[16:28] = 0.0050062894680040258624242888174649
            w[28:32] = 0.021428571428571428571428571428571
        else:
            raise ValueError("Dimension not supported")
    elif degree == 4:
        if sd == 2:
            alpha = 0.2113248654051871  # 0.2113248654051871
            betas = [
                0.4247639617258106,
                0.130791593829745
            ]
            x = list(ref_el.vertices)
            for e in range(3):
                x.extend(ref_el.make_points(1, e, 2))  # edge midpoints
            x.extend(
                [
                    (1 - alpha, alpha),
                    (alpha, 1 - alpha),
                    (0.0, 1 - alpha),
                    (0.0, alpha),
                    (alpha, 0.0),
                    (1 - alpha, 0.0),
                ]  # edge points
            )
            for beta in betas:
                x.extend(
                    [(beta, beta), (1 - 2 * beta, beta), (beta, 1 - 2 * beta)]
                )  # points in center of cell
            w = numpy.arange(18, dtype=numpy.float64)
            w[0:3] = 0.003174603174603175
            w[3:6] = 0.0126984126984127
            w[6:12] = 0.01071428571428571
            w[12:15] = 0.07878121446939182
            w[15:18] = 0.05058386489568756
        else:
            raise ValueError("Dimension not supported")

    elif degree == 5:
        if sd == 2:
            alphas = [
                0.3632980741536860e-00,
                0.1322645816327140e-00
            ]
            betas = [
                0.4578368380791611e-00,
                0.2568591072619591e-00,
                0.5752768441141011e-01
            ]
            gamma = 0.7819258362551702e-01
            delta = 0.2210012187598900e-00
            x = list(ref_el.vertices)
            for alpha in alphas:
                x.extend(
                    [
                        (1 - alpha, alpha),
                        (alpha, 1 - alpha),
                        (0.0, 1 - alpha),
                        (0.0, alpha),
                        (alpha, 0.0),
                        (1 - alpha, 0.0),
                    ]  # edge points
                )
            for beta in betas:
                x.extend(
                    [(beta, beta), (1 - 2 * beta, beta), (beta, 1 - 2 * beta)]
                )  # points in center of cell
            x.extend(
                [
                    (gamma, delta),
                    (1 - gamma - delta, delta),
                    (gamma, 1 - gamma - delta),
                    (delta, gamma),
                    (1 - gamma - delta, gamma),
                    (delta, 1 - gamma - delta),
                ]  # edge points
            )
            w = numpy.arange(30, dtype=numpy.float64)
            w[0:3] = 0.7094239706792450e-03
            w[3:9] = 0.6190565003676629e-02
            w[9:15] = 0.3480578640489211e-02
            w[15:18] = 0.3453043037728279e-01
            w[18:21] = 0.4590123763076286e-01
            w[21:24] = 0.1162613545961757e-01
            w[24:30] = 0.2727857596999626e-01
        else:
            raise ValueError("Dimension not supported")

    elif degree == 6:
        if sd == 2:
            x = list(ref_el.vertices)
            episilon = 5.00000000000000e-1
            alphas = [
                8.29411811106452e-2,
                2.68649695592714e-1,
            ]
            betas = [
                4.68059729056814e-1,
                7.93088545089875e-2,
                3.92931636618867e-1,
            ]
            gammas = [
                2.48172758709406e-1,
                1.56582066033687e-1,
            ]
            deltas = [
                6.99812197147049e-1,
                2.43089592364562e-1,
            ]
            weights = [
                5.35113520281665e-4,
                4.29435346026293e-3,
                3.02990950926060e-3,
                3.16396316646563e-3,
                2.43035184285235e-2,
                1.66312091329395e-2,
                3.42178857644876e-2,
                1.73480160090330e-2,
                1.98004044953264e-2,
            ]

            x.extend(
                [
                    (episilon, episilon),
                    (0.0, episilon),
                    (episilon, 0.0),
                ]
            )  # edge midpoints (class 2 points)
            for alpha in alphas:
                x.extend(
                    [
                        (1 - alpha, alpha),
                        (alpha, 1 - alpha),
                        (0.0, 1 - alpha),
                        (0.0, alpha),
                        (alpha, 0.0),
                        (1 - alpha, 0.0),
                    ]  # edge points (sets of class 3 points)
                )

            for beta in betas:
                x.extend(
                    [
                        (beta, beta),
                        (1 - 2 * beta, beta),
                        (beta, 1 - 2 * beta)
                    ]  # interior points on bisector (sets of class 5 points)
                )

            for gamma, delta in zip(gammas, deltas):
                x.extend(
                    [
                        (gamma, delta),
                        (1 - gamma - delta, delta),
                        (gamma, 1 - gamma - delta),
                        (delta, gamma),
                        (1 - gamma - delta, gamma),
                        (delta, 1 - gamma - delta),
                    ]  # interior points (sets of class 6 points)
                )

            w = numpy.arange(39, dtype=numpy.float64)
            w[0:3] = weights[0]  # class 1 points (vertices)
            w[3:6] = weights[1]  # class 2 points
            w[6:12] = weights[2]  # class 3 points
            w[12:18] = weights[3]
            w[18:21] = weights[4]  # class 5 points
            w[21:24] = weights[5]
            w[24:27] = weights[6]
            w[27:33] = weights[7]  # class 6 points
            w[33:39] = weights[8]

        else:
            raise ValueError("Dimension not supported")
    # Return scheme
    return QuadratureRule(T, x, w)


def xg_scheme(ref_el, degree):
    """A (nearly) Gaussian simplicial quadrature with very few quadrature nodes,
    available for low-to-moderate orders.

    Raises `ValueError` if no quadrature rule for the requested parameters is available.

    See

        H. Xiao and Z. Gimbutas, "A numerical algorithm for the construction of
        efficient quadrature rules in two and higher dimensions," Computers &
        Mathematics with Applications, vol. 59, no. 2, pp. 663-676, 2010.
        http://dx.doi.org/10.1016/j.camwa.2009.10.027
    """
    dim = ref_el.get_spatial_dimension()
    if dim == 2:
        from FIAT.xg_quad_data import triangle_table as table
    elif dim == 3:
        from FIAT.xg_quad_data import tetrahedron_table as table
    else:
        raise ValueError(f"Xiao-Gambutas rule not availale for {dim} dimensions.")
    try:
        order_table = table[degree]
    except KeyError:
        raise ValueError(f"Xiao-Gambutas rule not availale for degree {degree}.")

    pts_ref = order_table["points"]
    wts_ref = order_table["weights"]
    Ref1 = symmetric_simplex(dim)
    pts, wts = map_quadrature(pts_ref, wts_ref, Ref1, ref_el)
    return QuadratureRule(ref_el, pts, wts)


def _triangle_scheme(ref_el, degree):
    """Return a quadrature scheme on a triangle of specified order. Falls
    back on canonical rule for higher orders."""
    if degree == 0 or degree == 1:
        # Scheme from Zienkiewicz and Taylor, 1 point, degree of precision 1
        x = numpy.array([[1.0/3.0, 1.0/3.0]])
        w = numpy.array([0.5])
    elif degree == 2:
        # Scheme from Strang and Fix, 3 points, degree of precision 2
        x = numpy.array([[1.0/6.0, 1.0/6.0],
                         [1.0/6.0, 2.0/3.0],
                         [2.0/3.0, 1.0/6.0]])
        w = numpy.zeros((3,), dtype=numpy.float64)
        w.fill(1.0/6.0)
    elif degree == 3:
        # Scheme from Strang and Fix, 6 points, degree of precision 3
        x = numpy.array([[0.659027622374092, 0.231933368553031],
                         [0.659027622374092, 0.109039009072877],
                         [0.231933368553031, 0.659027622374092],
                         [0.231933368553031, 0.109039009072877],
                         [0.109039009072877, 0.659027622374092],
                         [0.109039009072877, 0.231933368553031]])
        w = numpy.zeros((6,), dtype=numpy.float64)
        w.fill(1.0/12.0)
    else:
        try:
            # Get Xiao-Gambutas scheme
            return xg_scheme(ref_el, degree)
        except ValueError:
            # Get canonical scheme
            return _fiat_scheme(ref_el, degree)

    # Return scheme
    x, w = map_quadrature(x, w, UFCTriangle(), ref_el)
    return QuadratureRule(ref_el, x, w)


def _tetrahedron_scheme(ref_el, degree):
    """Return a quadrature scheme on a tetrahedron of specified
    degree. Falls back on canonical rule for higher orders"""
    if degree == 0 or degree == 1:
        # Scheme from Zienkiewicz and Taylor, 1 point, degree of precision 1
        x = numpy.array([[1.0/4.0, 1.0/4.0, 1.0/4.0]])
        w = numpy.array([1.0/6.0])
    elif degree == 2:
        # Scheme from Zienkiewicz and Taylor, 4 points, degree of precision 2
        a, b = 0.585410196624969, 0.138196601125011
        x = numpy.array([[a, b, b],
                         [b, a, b],
                         [b, b, a],
                         [b, b, b]])
        w = numpy.zeros((4,), dtype=numpy.float64)
        w.fill(1.0/24.0)
    else:
        try:
            # Get Xiao-Gambutas scheme
            return xg_scheme(ref_el, degree)
        except ValueError:
            # Get canonical scheme
            return _fiat_scheme(ref_el, degree)

    # Return scheme
    x, w = map_quadrature(x, w, UFCTetrahedron(), ref_el)
    return QuadratureRule(ref_el, x, w)



# Copyright (C) 2025 Imperial College London and others
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#
# Pablo Brubeck

import pytest
import numpy as np

from FIAT import ufc_simplex, Histopolation, GaussLobattoLegendre
from FIAT.barycentric_interpolation import get_lagrange_points
from FIAT.macro import IsoSplit
from FIAT.quadrature_schemes import create_quadrature


@pytest.fixture
def interval():
    return ufc_simplex(1)


@pytest.mark.parametrize("degree", range(7))
def test_gll_basis_values(interval, degree):
    """Ensure that integrating a simple monomial produces the expected results."""

    s = interval
    q = create_quadrature(s, 2*degree)
    fe = Histopolation(s, degree)
    tab = fe.tabulate(0, q.pts)[(0,)]

    for test_degree in range(degree + 1):
        v = lambda x: sum(x)**test_degree
        coefs = [n(v) for n in fe.dual.nodes]
        integral = np.dot(coefs, np.dot(tab, q.wts))
        reference = q.integrate(v)
        assert np.allclose(integral, reference, rtol=1e-14)


@pytest.mark.parametrize("degree", range(4))
def test_histopolation_dofs(interval, degree):
    """Ensure that basis functions integrate to either 1 or 0 on the GLL subgrid."""

    fe = Histopolation(interval, degree)
    gll = GaussLobattoLegendre(interval, degree+1)
    pts = get_lagrange_points(gll.dual_basis())
    x = np.reshape(pts, (-1, ))

    macrocell = IsoSplit(interval, degree+1, variant="gll")
    assert np.allclose(macrocell.vertices, pts)

    Q = create_quadrature(macrocell, 2*degree)
    qpts = Q.get_points()
    qwts = Q.get_weights().reshape((degree+1, -1))

    tab = fe.tabulate(0, qpts)[(0,)]
    tab = tab.reshape((fe.space_dimension(), degree+1, -1))
    expected = np.eye(degree+1)
    for i in range(degree+1):
        result = np.dot(tab[:, i, :], qwts[i] / (x[i+1] - x[i]))
        assert np.allclose(result, expected[i])



from itertools import chain, combinations

import numpy

from FIAT import expansions, polynomial_set, reference_element
from FIAT.quadrature import FacetQuadratureRule, QuadratureRule
from FIAT.reference_element import (TRIANGLE, SimplicialComplex, lattice_iter,
                                    make_lattice)


def bary_to_xy(verts, bary, result=None):
    """Maps barycentric coordinates to physical points.

    :arg verts: A tuple of points.
    :arg bary: A row-stacked numpy array of barycentric coordinates.
    :arg result: A row-stacked numpy array of physical points.
    :returns: result
    """
    return numpy.dot(bary, verts, out=result)


def xy_to_bary(verts, pts, result=None):
    """Maps physical points to barycentric coordinates.

    :arg verts: A tuple of points.
    :arg pts: A row-stacked numpy array of physical points.
    :arg result: A row-stacked numpy array of barycentric coordinates.
    :returns: result
    """
    verts = numpy.asarray(verts)
    pts = numpy.asarray(pts)
    npts = pts.shape[0]
    sdim = verts.shape[1]

    mat = numpy.vstack((verts.T, numpy.ones((1, sdim+1))))
    b = numpy.vstack((pts.T, numpy.ones((1, npts))))
    foo = numpy.linalg.solve(mat, b)
    if result is None:
        return numpy.copy(foo.T)
    else:
        result[:, :] = foo[:, :].T
    return result


def facet_support(facet_coords, tol=1.e-12):
    """Returns the support of a facet.

    :arg facet_coords: An iterable of tuples (barycentric coordinates) describing the facet.
    :returns: A tuple of vertex ids where some coordinate is nonzero.
    """
    return tuple(sorted(set(i for x in facet_coords for (i, xi) in enumerate(x) if abs(xi) > tol)))


def invert_cell_topology(T):
    """Returns a dict of dicts mapping dimension x vertices to entity id."""
    return {dim: {T[dim][entity]: entity for entity in T[dim]} for dim in T}


def make_topology(sd, num_verts, edges):
    topology = {}
    topology[0] = {i: (i,) for i in range(num_verts)}
    topology[1] = dict(enumerate(sorted(edges)))

    # Get an adjacency list for each vertex
    adjacency = {v: set(chain.from_iterable(verts for verts in edges if v in verts))
                 for v in topology[0]}

    # Complete the higher dimensional facets by appending a vertex
    # adjacent to the vertices of codimension 1 facets
    for dim in range(1, sd):
        entities = []
        for entity in topology[dim]:
            facet = topology[dim][entity]
            facet_verts = set(facet)
            for v in range(min(facet)):
                if (facet_verts < adjacency[v]):
                    entities.append((v, *facet))

        topology[dim+1] = dict(enumerate(sorted(entities)))
    return topology


class SplitSimplicialComplex(SimplicialComplex):
    """Abstract class to implement a split on a Simplex.

    :arg parent: The parent Simplex to split.
    :arg vertices: The vertices of the simplicial complex.
    :arg topology: The topology of the simplicial complex.
    """
    def __init__(self, parent, vertices, topology):
        self._parent_complex = parent
        while parent.get_parent():
            parent = parent.get_parent()
        self._parent_simplex = parent

        bary = xy_to_bary(parent.get_vertices(), vertices)
        parent_top = parent.get_topology()
        parent_inv_top = invert_cell_topology(parent_top)

        # dict mapping child facets to their parent facet
        child_to_parent = {}
        # dict mapping parent facets to their children
        parent_to_children = {dim: {entity: [] for entity in parent_top[dim]} for dim in parent_top}
        for dim in topology:
            child_to_parent[dim] = {}
            for entity in topology[dim]:
                facet_ids = topology[dim][entity]
                facet_coords = bary[list(facet_ids), :]
                parent_verts = facet_support(facet_coords)
                parent_dim = len(parent_verts) - 1
                parent_entity = parent_inv_top[parent_dim][parent_verts]
                child_to_parent[dim][entity] = (parent_dim, parent_entity)
                parent_to_children[parent_dim][parent_entity].append((dim, entity))

        for dim in parent_to_children:
            for entity in parent_to_children[dim]:
                children = parent_to_children[dim][entity]
                if len(children) > 1:
                    # sort children lexicographically
                    pts = [tuple(numpy.average([vertices[i] for i in topology[cdim][centity]], 0))
                           for cdim, centity in children]
                    bary = parent.compute_barycentric_coordinates(pts, entity=(dim, entity))
                    order = numpy.lexsort(bary.T)
                    children = tuple(children[j] for j in order)
                else:
                    children = tuple(children)
                parent_to_children[dim][entity] = children

        self._child_to_parent = child_to_parent
        self._parent_to_children = parent_to_children

        sd = parent.get_spatial_dimension()
        inv_top = invert_cell_topology(topology)

        # dict mapping cells to their boundary facets for each dimension,
        # while respecting the ordering on the parent simplex
        connectivity = {cell: {dim: [] for dim in topology} for cell in topology[sd]}
        for cell in topology[sd]:
            cell_verts = topology[sd][cell]
            for dim in parent_top:
                for entity in parent_top[dim]:
                    ref_verts = parent_top[dim][entity]
                    global_verts = tuple(cell_verts[v] for v in ref_verts)
                    connectivity[cell][dim].append(inv_top[dim][global_verts])
        self._cell_connectivity = connectivity

        # dict mapping subentity dimension to interior facets
        interior_facets = {dim: [entity for entity in child_to_parent[dim]
                                 if child_to_parent[dim][entity][0] == sd]
                           for dim in sorted(child_to_parent)}
        self._interior_facets = interior_facets

        super().__init__(parent.shape, vertices, topology)

    def get_child_to_parent(self):
        """Maps split complex facet tuple to its parent entity tuple."""
        return self._child_to_parent

    def get_parent_to_children(self):
        """Maps parent facet tuple to a list of tuples of entites in the split complex."""
        return self._parent_to_children

    def get_cell_connectivity(self):
        """Connectitivity from cell in a complex to global facet ids and
        respects the entity numbering on the reference cell.

        N.B. cell_connectivity[cell][dim] has the same contents as
        self.connectivity[(sd, dim)][cell], except those are sorted.
        """
        return self._cell_connectivity

    def get_interior_facets(self, dimension):
        """Returns the list of entities of the given dimension that are
        supported on the parent's interior.

        :arg dimension: subentity dimension (integer)
        """
        return self._interior_facets[dimension]

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        return self.get_parent().construct_subelement(dimension)

    def get_facet_element(self):
        dimension = self.get_spatial_dimension()
        return self.construct_subelement(dimension - 1)

    def is_macrocell(self):
        return True

    def get_parent(self):
        return self._parent_simplex

    def get_parent_complex(self):
        return self._parent_complex


class IsoSplit(SplitSimplicialComplex):
    """Splits simplex into the simplicial complex obtained by
    connecting points on a regular lattice.

    :arg ref_el: The parent Simplex to split.
    :kwarg degree: The number of subdivisions along each edge of the simplex.
    :kwarg variant: The point distribution variant.
    """
    def __init__(self, ref_el, degree=2, variant=None):
        self.degree = degree
        self.variant = variant
        # Place new vertices on a lattice
        sd = ref_el.get_spatial_dimension()
        new_verts = make_lattice(ref_el.vertices, degree, variant=variant)
        flat_index = {tuple(alpha): i for i, alpha in enumerate(lattice_iter(0, degree+1, sd))}

        # Loop through degree-1 vertices
        # Construct a P1 simplex by connecting edges between a vertex and
        # its neighbors obtained by shifting each coordinate up by 1
        edges = []
        for alpha in lattice_iter(0, degree, sd):
            simplex = []
            for beta in lattice_iter(0, 2, sd):
                v1 = flat_index[tuple(a+b for a, b in zip(alpha, beta))]
                edges.extend((v0, v1) for v0 in simplex)
                simplex.append(v1)

        if sd == 3:
            # Cut the octahedron
            # FIXME do this more generically
            assert degree == 2
            v0, v1 = flat_index[(1, 0, 0)], flat_index[(0, 1, 1)]
            edges.append(tuple(sorted((v0, v1))))

        new_topology = make_topology(sd, len(new_verts), edges)
        super().__init__(ref_el, tuple(new_verts), new_topology)

    def construct_subcomplex(self, dimension):
        """Constructs the reference subcomplex of the parent complex
        specified by subcomplex dimension.
        """
        if dimension == self.get_dimension():
            return self
        ref_el = self.construct_subelement(dimension)
        if dimension == 0:
            return ref_el
        else:
            # Iso on facets is Iso
            return IsoSplit(ref_el, self.degree, self.variant)


class PowellSabinSplit(SplitSimplicialComplex):
    """Splits a simplicial complex by connecting barycenters of subentities
    to the barycenter of the superentities of the given dimension or higher.
    """
    def __init__(self, ref_el, dimension=1):
        self.split_dimension = dimension
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()
        connectivity = ref_el.get_connectivity()
        new_verts = list(ref_el.get_vertices())
        dim = dimension - 1
        simplices = {dim: {entity: [top[dim][entity]] for entity in top[dim]}}

        for dim in range(dimension, sd+1):
            simplices[dim] = {}
            for entity in top[dim]:
                bary_id = len(new_verts)
                new_verts.extend(ref_el.make_points(dim, entity, dim+1))

                # Connect entity barycenter to every subsimplex on the entity
                simplices[dim][entity] = [(*s, bary_id)
                                          for child in connectivity[(dim, dim-1)][entity]
                                          for s in simplices[dim-1][child]]

        simplices = list(chain.from_iterable(simplices[sd].values()))
        new_topology = {}
        new_topology[0] = {i: (i,) for i in range(len(new_verts))}
        for dim in range(1, sd):
            facets = chain.from_iterable((combinations(s, dim+1) for s in simplices))
            if dim < self.split_dimension:
                # Preserve the numbering of the unsplit entities
                facets = chain(top[dim].values(), facets)
            unique_facets = dict.fromkeys(facets)
            new_topology[dim] = dict(enumerate(unique_facets))
        new_topology[sd] = dict(enumerate(simplices))

        parent = ref_el if dimension == sd else PowellSabinSplit(ref_el, dimension=dimension+1)
        super().__init__(parent, tuple(new_verts), new_topology)

    def construct_subcomplex(self, dimension):
        """Constructs the reference subcomplex of the parent complex
        specified by subcomplex dimension.
        """
        if dimension == self.get_dimension():
            return self
        parent = self.get_parent_complex()
        subcomplex = parent.construct_subcomplex(dimension)
        if dimension < self.split_dimension:
            return subcomplex
        else:
            # Powell-Sabin on facets is Powell-Sabin
            return PowellSabinSplit(subcomplex, dimension=self.split_dimension)


class AlfeldSplit(PowellSabinSplit):
    """Splits a simplicial complex by connecting cell vertices to their
    barycenter.
    """
    def __new__(cls, ref_el):
        try:
            return ref_el._split_cache[cls]
        except KeyError:
            self = super().__new__(cls)
            return ref_el._split_cache.setdefault(cls, self)

    def __init__(self, ref_el):
        sd = ref_el.get_spatial_dimension()
        super().__init__(ref_el, dimension=sd)


class WorseyFarinSplit(PowellSabinSplit):
    """Splits a simplicial complex by connecting cell and facet vertices to their
    barycenter. This reduces to Powell-Sabin on the triangle, and Alfeld on the interval.
    """
    def __new__(cls, ref_el):
        try:
            return ref_el._split_cache[cls]
        except KeyError:
            self = super().__new__(cls)
            return ref_el._split_cache.setdefault(cls, self)

    def __init__(self, ref_el):
        sd = ref_el.get_spatial_dimension()
        super().__init__(ref_el, dimension=sd-1)


class PowellSabin12Split(SplitSimplicialComplex):
    """Splits a triangle (only!) by connecting each vertex to the opposite
    edge midpoint and edge midpoints to each other.
    """
    def __init__(self, ref_el):
        assert ref_el.get_shape() == TRIANGLE
        verts = ref_el.get_vertices()
        new_verts = list(verts)
        new_verts.extend(
            map(tuple, bary_to_xy(verts,
                [(1/3, 1/3, 1/3),
                 (1/2, 1/2, 0),
                 (1/2, 0, 1/2),
                 (0, 1/2, 1/2),
                 (1/2, 1/4, 1/4),
                 (1/4, 1/2, 1/4),
                 (1/4, 1/4, 1/2)])))

        edges = [(0, 4), (0, 7), (0, 5),
                 (1, 4), (1, 8), (1, 6),
                 (2, 5), (2, 9), (2, 6),
                 (3, 4), (3, 5), (3, 6), (3, 7), (3, 8), (3, 9),
                 (4, 7), (4, 8), (5, 7), (5, 9), (6, 8), (6, 9)]

        parent = PowellSabinSplit(ref_el)
        new_topology = make_topology(2, len(new_verts), edges)
        super().__init__(parent, tuple(new_verts), new_topology)

    def construct_subcomplex(self, dimension):
        """Constructs the reference subcomplex of the parent cell subentity
        specified by subcomplex dimension.
        """
        if dimension == 2:
            return self
        elif dimension == 1:
            return AlfeldSplit(self.construct_subelement(1))
        elif dimension == 0:
            return self.construct_subelement(0)
        else:
            raise ValueError("Illegal dimension")


class MacroQuadratureRule(QuadratureRule):
    """Composite quadrature rule on parent facets that respects the splitting.

    :arg ref_el: A simplicial complex.
    :arg Q_ref: A QuadratureRule on the reference simplex.
    :args parent_facets: An iterable of facets of the same dimension as Q_ref,
                         defaults to all facets.

    :returns: A quadrature rule on the sub entities of the simplicial complex.
    """
    def __init__(self, ref_el, Q_ref, parent_facets=None):
        parent_dim = Q_ref.ref_el.get_spatial_dimension()
        if parent_facets is not None:
            parent_to_children = ref_el.get_parent_to_children()
            facets = []
            for parent_entity in parent_facets:
                children = parent_to_children[parent_dim][parent_entity]
                facets.extend(entity for dim, entity in children if dim == parent_dim)
        else:
            top = ref_el.get_topology()
            facets = top[parent_dim]

        pts = []
        wts = []
        for entity in facets:
            Q_cur = FacetQuadratureRule(ref_el, parent_dim, entity, Q_ref)
            pts.extend(Q_cur.pts)
            wts.extend(Q_cur.wts)
        pts = tuple(pts)
        wts = tuple(wts)
        super().__init__(ref_el, pts, wts)


class CkPolynomialSet(polynomial_set.PolynomialSet):
    """Constructs a C^k-continuous PolynomialSet on a simplicial complex.

    :arg ref_el: The simplicial complex.
    :arg degree: The polynomial degree.
    :kwarg order: The order of continuity across facets or a dict of dicts
        mapping dimension and entity_id to the order of continuity at each entity.
    :kwarg vorder: The order of super-smoothness at interior vertices.
    :kwarg shape: The value shape.
    :kwarg variant: The variant for the underlying ExpansionSet.
    :kwarg scale: The scale for the underlying ExpansionSet.
    """
    def __init__(self, ref_el, degree, order=1, vorder=None, shape=(), **kwargs):
        from FIAT.quadrature_schemes import create_quadrature

        if not isinstance(order, (int, dict)):
            raise TypeError(f"'order' must be either an int or dict, not {type(order).__name__}")

        sd = ref_el.get_spatial_dimension()
        if isinstance(order, int):
            order = {sd-1: dict.fromkeys(ref_el.get_interior_facets(sd-1), order)}
        if vorder is not None:
            order[0] = dict.fromkeys(ref_el.get_interior_facets(0), vorder)
        elif 0 not in order:
            order[0] = {}

        if not all(k in {0, sd-1} for k in order):
            raise NotImplementedError("Only face or vertex constraints have been implemented.")

        expansion_set = expansions.ExpansionSet(ref_el, **kwargs)
        k = 1 if expansion_set.continuity == "C0" else 0

        # Impose C^forder continuity across interior facets
        facet_el = ref_el.construct_subelement(sd-1)
        phi_deg = 0 if sd == 1 else degree - k
        phi = polynomial_set.ONPolynomialSet(facet_el, phi_deg)
        Q = create_quadrature(facet_el, 2 * phi_deg)
        qpts, qwts = Q.get_points(), Q.get_weights()
        phi_at_qpts = phi.tabulate(qpts)[(0,) * (sd-1)]
        weights = numpy.multiply(phi_at_qpts, qwts)

        rows = []
        for facet in order[sd-1]:
            forder = order[sd-1][facet]
            jumps = expansion_set.tabulate_normal_jumps(degree, qpts, facet, order=forder)
            for r in range(k, forder+1):
                num_wt = 1 if sd == 1 else expansions.polynomial_dimension(facet_el, degree-r)
                rows.append(numpy.tensordot(weights[:num_wt], jumps[r], axes=(-1, -1)))

        # Impose C^vorder super-smoothness at interior vertices
        # C^forder automatically gives C^{forder+dim-1} at the interior vertex
        verts = numpy.asarray(ref_el.get_vertices())
        for vorder in set(order[0].values()):
            vids = [i for i in order[0] if order[0][i] == vorder]
            facets = chain.from_iterable(ref_el.connectivity[(0, sd-1)][v] for v in vids)
            forder = min(order[sd-1][f] for f in facets)
            sorder = forder + sd - 1
            if vorder > sorder:
                jumps = expansion_set.tabulate_jumps(degree, verts[vids], order=vorder)
                rows.extend(numpy.vstack(jumps[r].T) for r in range(sorder+1, vorder+1))

        if len(rows) > 0:
            for row in rows:
                row *= 1 / max(numpy.max(abs(row)), 1)
            dual_mat = numpy.vstack(rows)
            coeffs = polynomial_set.spanning_basis(dual_mat, nullspace=True)
        else:
            coeffs = numpy.eye(expansion_set.get_num_members(degree))

        if shape != ():
            m, n = coeffs.shape
            ncomp = numpy.prod(shape)
            coeffs = numpy.kron(coeffs, numpy.eye(ncomp))
            coeffs = coeffs.reshape(m*ncomp, *shape, n)

        super().__init__(ref_el, degree, degree, expansion_set, coeffs)


def hdiv_conforming_coefficients(U, order=0):
    """Constrains a PolynomialSet to have vanishing normal jumps on the
       interior facets of a simplicial complex.

    :arg U: A PolynomialSet, it may be vector- or tensor-valued.
    :arg order: The maximum order of differentiation on the normal jump constraints.

    :returns: The coefficients of the constrained PolynomialSet.
    """
    from FIAT.quadrature_schemes import create_quadrature
    degree = U.degree
    ref_el = U.get_reference_element()
    coeffs = U.get_coeffs()
    shape = U.get_shape()
    expansion_set = U.get_expansion_set()
    k = 1 if expansion_set.continuity == "C0" else 0

    sd = ref_el.get_spatial_dimension()
    facet_el = ref_el.construct_subelement(sd-1)

    phi_deg = 0 if sd == 1 else degree - k
    phi = polynomial_set.ONPolynomialSet(facet_el, phi_deg, shape=shape[1:])
    Q = create_quadrature(facet_el, 2 * phi_deg)
    qpts, qwts = Q.get_points(), Q.get_weights()
    phi_at_qpts = phi.tabulate(qpts)[(0,) * (sd-1)]
    weights = numpy.multiply(phi_at_qpts, qwts)
    ax = tuple(range(1, weights.ndim))

    rows = []
    for facet in ref_el.get_interior_facets(sd-1):
        normal = ref_el.compute_scaled_normal(facet)
        ncoeffs = numpy.tensordot(coeffs, normal, axes=(len(shape), 0))
        jumps = expansion_set.tabulate_normal_jumps(degree, qpts, facet, order=order)
        for r in range(k, order+1):
            njump = numpy.dot(ncoeffs, jumps[r])
            rows.append(numpy.tensordot(weights, njump, axes=(ax, ax)))

    if len(rows) > 0:
        dual_mat = numpy.vstack(rows)
        nsp = polynomial_set.spanning_basis(dual_mat, nullspace=True)
        coeffs = numpy.tensordot(nsp, coeffs, axes=(1, 0))
    return coeffs


class HDivPolynomialSet(polynomial_set.PolynomialSet):
    """Constructs a vector-valued PolynomialSet with continuous
       normal components on a simplicial complex.

    :arg ref_el: The simplicial complex.
    :arg degree: The polynomial degree.
    :kwarg order: The order of continuity across subcells.
    :kwarg variant: The variant for the underlying ExpansionSet.
    :kwarg scale: The scale for the underlying ExpansionSet.
    """
    def __init__(self, ref_el, degree, order=0, **kwargs):
        sd = ref_el.get_spatial_dimension()
        U = polynomial_set.ONPolynomialSet(ref_el, degree, shape=(sd,), **kwargs)
        coeffs = hdiv_conforming_coefficients(U, order=order)
        super().__init__(ref_el, degree, degree, U.expansion_set, coeffs)


class HDivSymPolynomialSet(polynomial_set.PolynomialSet):
    """Constructs a symmetric tensor-valued PolynomialSet with continuous
       normal components on a simplicial complex.

    :arg ref_el: The simplicial complex.
    :arg degree: The polynomial degree.
    :kwarg order: The order of continuity across subcells.
    :kwarg variant: The variant for the underlying ExpansionSet.
    :kwarg scale: The scale for the underlying ExpansionSet.
    """
    def __init__(self, ref_el, degree, order=0, **kwargs):
        U = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree, **kwargs)
        coeffs = hdiv_conforming_coefficients(U, order=order)
        super().__init__(ref_el, degree, degree, U.expansion_set, coeffs)


def pullback(phi, mapping, J=None, Jinv=None, Jdet=None):
    """Transforms a reference tabulation into physical space.

    :arg phi: The tabulation on reference space.
    :arg mapping: A string indicating the pullback type.
    :arg J: The Jacobian of the transformation from reference to physical space.
    :arg Jinv: The inverse of the Jacobian.
    :arg Jdet: The determinant of the Jacobian.

    :returns: The tabulation on physical space.
    """
    try:
        formdegree = {
            "affine": (0,),
            "covariant piola": (1,),
            "contravariant piola": (2,),
            "double covariant piola": (1, 1),
            "double contravariant piola": (2, 2),
            "covariant contravariant piola": (1, 2),
            "contravariant covariant piola": (2, 1), }[mapping]
    except KeyError:
        raise ValueError(f"Unrecognized mapping {mapping}")

    if J is None:
        J = numpy.linalg.pinv(Jinv)
    if Jinv is None:
        Jinv = numpy.linalg.pinv(J)
    if Jdet is None:
        Jdet = numpy.linalg.det(J)
    F1 = Jinv.T
    F2 = J / Jdet

    for i, k in enumerate(formdegree):
        if k == 0:
            continue
        F = F1 if k == 1 else F2

        perm = list(range(phi.ndim))
        perm[i+1], perm[-1] = perm[-1], perm[i+1]

        phi = phi.transpose(perm)
        phi = phi.dot(F.T)
        phi = phi.transpose(perm)

    return phi


class MacroPolynomialSet(polynomial_set.PolynomialSet):
    """Constructs a PolynomialSet by tiling a CiarletElement on a
       SimplicialComplex.

    :arg ref_el: The SimplicialComplex.
    :arg element: The CiarletElement.
    """
    def __init__(self, ref_el, element):
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()

        mapping, = set(element.mapping())
        base_ref_el = element.get_reference_element()
        base_entity_ids = element.entity_dofs()
        n = element.degree()

        base_expansion_set = element.get_nodal_basis().get_expansion_set()
        expansion_set = base_expansion_set.reconstruct(ref_el=ref_el)

        shp = element.value_shape()
        num_bfs = expansions.polynomial_dimension(ref_el, n, base_entity_ids)
        num_members = expansion_set.get_num_members(n)
        coeffs = numpy.zeros((num_bfs, *shp, num_members))
        base_coeffs = element.get_coeffs()

        rmap = expansions.polynomial_cell_node_map(ref_el, n, base_entity_ids)
        cmap = expansion_set.get_cell_node_map(n)
        for cell in sorted(top[sd]):
            cell_verts = ref_el.get_vertices_of_subcomplex(top[sd][cell])
            A, b = reference_element.make_affine_mapping(base_ref_el.vertices, cell_verts)

            indices = numpy.ix_(rmap[cell], *map(range, shp), cmap[cell])
            coeffs[indices] = pullback(base_coeffs, mapping, J=A)

        super().__init__(ref_el, element.degree(), element.degree(), expansion_set, coeffs)



from finat.point_set import UnknownPointSet, FacetPointSet

import numpy

import FIAT

import gem
from gem.interpreter import evaluate
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase
from finat.quadrature import make_quadrature, AbstractQuadratureRule


def make_quadrature_element(fiat_ref_cell, degree, scheme="default", codim=0):
    """Construct a :class:`QuadratureElement` from a given a reference
    element, degree and scheme.

    :param fiat_ref_cell: The FIAT reference cell to build the
        :class:`QuadratureElement` on.
    :param degree: The degree of polynomial that the rule should
        integrate exactly.
    :param scheme: The quadrature scheme to use - e.g. "default",
        "canonical" or "KMV".
    :param codim: The codimension of the quadrature scheme.
    :returns: The appropriate :class:`QuadratureElement`
    """
    if codim > 0:
        sd = fiat_ref_cell.get_spatial_dimension()
        rule_ref_cell = fiat_ref_cell.construct_subelement(sd - codim)
    else:
        rule_ref_cell = fiat_ref_cell

    rule = make_quadrature(rule_ref_cell, degree, scheme=scheme)
    return QuadratureElement(fiat_ref_cell, rule)


class QuadratureElement(FiniteElementBase):
    """A set of quadrature points pretending to be a finite element."""

    def __init__(self, fiat_ref_cell, rule):
        """Construct a :class:`QuadratureElement`.

        :param fiat_ref_cell: The FIAT reference cell to build the
            :class:`QuadratureElement` on
        :param rule: A :class:`AbstractQuadratureRule` to use
        """
        self.cell = fiat_ref_cell
        if not isinstance(rule, AbstractQuadratureRule):
            raise TypeError("rule is not an AbstractQuadratureRule")
        self._rule = rule

    @cached_property
    def cell(self):
        pass  # set at initialisation

    @property
    def complex(self):
        return self.cell

    @property
    def degree(self):
        raise NotImplementedError("QuadratureElement does not represent a polynomial space.")

    @property
    def formdegree(self):
        return None

    @cached_property
    def _entity_dofs(self):
        top = self.cell.get_topology()
        entity_dofs = {dim: {entity: [] for entity in entities}
                       for dim, entities in top.items()}
        ps = self._rule.point_set
        num_pts = len(ps.points)
        to_int = lambda x: sum(x) if isinstance(x, tuple) else x
        cur = 0
        for dim in sorted(top):
            if to_int(dim) == ps.dimension:
                for entity in sorted(top[dim]):
                    entity_dofs[dim][entity].extend(range(cur, cur + num_pts))
                    cur += num_pts
        return entity_dofs

    def entity_dofs(self):
        return self._entity_dofs

    def space_dimension(self):
        return numpy.prod(self.index_shape, dtype=int)

    @cached_property
    def _point_set(self):
        ps = self._rule.point_set
        sd = self.cell.get_spatial_dimension()
        return ps if ps.dimension == sd else FacetPointSet(self.cell, ps)

    @property
    def index_shape(self):
        ps = self._point_set
        return tuple(index.extent for index in ps.indices)

    @property
    def value_shape(self):
        return ()

    @cached_property
    def fiat_equivalent(self):
        ps = self._point_set
        if isinstance(ps, UnknownPointSet):
            raise ValueError("A quadrature element with rule with runtime points has no fiat equivalent!")
        weights = getattr(self._rule, 'weights', None)
        if weights is None:
            # we need the weights.
            weights, = evaluate([self._rule.weight_expression])
            weights = weights.arr.flatten()
            self._rule.weights = weights

        return FIAT.QuadratureElement(self.cell, ps.points, weights)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set object.
        :param entity: the cell entity on which to tabulate.
        '''
        rule_dim = self._rule.point_set.dimension
        if entity is None:
            entity = (rule_dim, 0)
        entity_dim, entity_id = entity
        if entity_dim != rule_dim:
            raise ValueError(f"Cannot tabulate QuadratureElement of dimension {rule_dim}"
                             f" on subentities of dimension {entity_dim}.")

        if order:
            raise ValueError("Derivatives are not defined on a QuadratureElement.")

        if not self._rule.point_set.almost_equal(ps):
            raise ValueError("Mismatch of quadrature points!")

        # Return an outer product of identity matrices
        basis_indices = self.get_indices()
        point_indices = ps.indices
        if len(basis_indices) > len(point_indices):
            point_indices = (entity_id, *point_indices)
        delta = gem.Delta(point_indices, basis_indices)

        sd = self.cell.get_spatial_dimension()
        return {(0,) * sd: gem.ComponentTensor(delta, basis_indices)}

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        raise NotImplementedError("QuadratureElement cannot do point evaluation!")

    @property
    def dual_basis(self):
        ps = self._point_set
        multiindex = self.get_indices()
        # Evaluation matrix is just an outer product of identity
        # matrices, evaluation points are just the quadrature points.
        Q = gem.Delta(ps.indices, multiindex)
        Q = gem.ComponentTensor(Q, multiindex)
        return Q, ps

    @property
    def mapping(self):
        return "affine"



import numpy
import pytest
from FIAT import QuadraticPowellSabin6 as PS6
from FIAT import QuadraticPowellSabin12 as PS12
from FIAT.functional import PointEvaluation
from FIAT.reference_element import make_lattice, ufc_simplex


@pytest.fixture
def cell():
    return ufc_simplex(2)


@pytest.mark.parametrize("el", (PS6, PS12))
def test_powell_sabin_constant(cell, el):
    # Test that bfs associated with point evaluation sum up to 1
    fe = el(cell)

    pts = make_lattice(cell.get_vertices(), 3)
    tab = fe.tabulate(2, pts)

    coeffs = numpy.zeros((fe.space_dimension(),))
    nodes = fe.dual_basis()
    entity_dofs = fe.entity_dofs()
    for v in entity_dofs[0]:
        for k in entity_dofs[0][v]:
            if isinstance(nodes[k], PointEvaluation):
                coeffs[k] = 1.0

    for alpha in tab:
        expected = 1 if sum(alpha) == 0 else 0
        assert numpy.allclose(coeffs @ tab[alpha], expected)



import numpy

import gem

from finat.finiteelementbase import FiniteElementBase
from finat.enriched import EnrichedElement


def MixedElement(elements):
    """Constructor function for FEniCS-style mixed elements.

    Implements mixed element using :py:class:`EnrichedElement` and
    value shape transformations with :py:class:`MixedSubElement`.
    """
    sizes = [numpy.prod(element.value_shape, dtype=int)
             for element in elements]
    offsets = [int(offset) for offset in numpy.cumsum([0] + sizes)]
    total_size = offsets.pop()
    return EnrichedElement([MixedSubElement(element, total_size, offset)
                            for offset, element in zip(offsets, elements)])


class MixedSubElement(FiniteElementBase):
    """Element wrapper that flattens value shape and places the flattened
    vector in a longer vector of zeros."""

    def __init__(self, element, size, offset):
        assert 0 <= offset <= size
        assert offset + numpy.prod(element.value_shape, dtype=int) <= size

        super().__init__()
        self.element = element
        self.size = size
        self.offset = offset

    @property
    def cell(self):
        return self.element.cell

    @property
    def complex(self):
        return self.element.complex

    @property
    def degree(self):
        return self.element.degree

    @property
    def formdegree(self):
        return self.element.formdegree

    def entity_dofs(self):
        return self.element.entity_dofs()

    def entity_closure_dofs(self):
        return self.element.entity_closure_dofs()

    def entity_support_dofs(self):
        return self.element.entity_support_dofs()

    def space_dimension(self):
        return self.element.space_dimension()

    @property
    def index_shape(self):
        return self.element.index_shape

    @property
    def value_shape(self):
        return (self.size,)

    def _transform(self, v):
        u = [gem.Zero()] * self.size
        for j, zeta in enumerate(numpy.ndindex(self.element.value_shape)):
            u[self.offset + j] = gem.Indexed(v, zeta)
        return u

    def _transform_evaluation(self, core_eval):
        beta = self.get_indices()
        zeta = self.get_value_indices()

        def promote(table):
            v = gem.partial_indexed(table, beta)
            u = gem.ListTensor(self._transform(v))
            return gem.ComponentTensor(gem.Indexed(u, zeta), beta + zeta)

        return {alpha: promote(table)
                for alpha, table in core_eval.items()}

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        core_eval = self.element.basis_evaluation(order, ps, entity, coordinate_mapping=coordinate_mapping)
        return self._transform_evaluation(core_eval)

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        core_eval = self.element.point_evaluation(order, refcoords, entity)
        return self._transform_evaluation(core_eval)

    @property
    def mapping(self):
        return self.element.mapping



import FIAT
from gem import ListTensor

from finat.aw import _facet_transform
from finat.fiat_elements import FiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement


class JohnsonMercier(PhysicallyMappedElement, FiatElement):  # symmetric matrix valued
    def __init__(self, cell, degree=1, variant=None):
        if Citations is not None:
            Citations().register("Gopalakrishnan2024")
        self._indices = slice(None, None)
        super().__init__(FIAT.JohnsonMercier(cell, degree, variant=variant))

    def basis_transformation(self, coordinate_mapping):
        numbf = self._element.space_dimension()
        ndof = self.space_dimension()

        V = identity(numbf, ndof)
        Vsub = _facet_transform(self.cell, 1, coordinate_mapping)
        Vsub = Vsub[:, self._indices]
        m, n = Vsub.shape
        V[:m, :n] = Vsub

        # Note: that the edge DOFs are scaled by edge lengths in FIAT implies
        # that they are already have the necessary rescaling to improve
        # conditioning.
        return ListTensor(V.T)



from FIAT import finite_element, dual_set, polynomial_set, expansions
from FIAT.check_format_variant import check_format_variant
from FIAT.functional import TensorBidirectionalIntegralMoment as BidirectionalMoment
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule
from FIAT.restricted import RestrictedElement


class GLSDual(dual_set.DualSet):
    def __init__(self, ref_el, degree):
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()
        nodes = []
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        # Face dofs: moments of normal-tangential components against a basis for P_k
        # Interior dofs: moments of normal-tangential components against a basis for P_{k-1}
        for dim in (sd-1, sd):
            q = degree + sd-1 - dim
            if q < 0:
                continue

            ref_facet = ref_el.construct_subelement(dim)
            Q_ref = create_quadrature(ref_facet, degree + q)
            P = polynomial_set.ONPolynomialSet(ref_facet, q, scale=1)
            phis = P.tabulate(Q_ref.get_points())[(0,) * dim]

            for entity in sorted(top[dim]):
                cur = len(nodes)
                Q = FacetQuadratureRule(ref_el, dim, entity, Q_ref)
                Jdet = Q.jacobian_determinant()
                for f in ref_el.get_connectivity()[(dim, sd-1)][entity]:
                    normal = ref_el.compute_scaled_normal(f)
                    tangents = ref_el.compute_tangents(sd-1, f)
                    n = normal / Jdet
                    nodes.extend(BidirectionalMoment(ref_el, t, n, Q, phi)
                                 for phi in phis for t in tangents)
                entity_ids[dim][entity].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class GopalakrishnanLedererSchoberlSecondKind(finite_element.CiarletElement):
    """The GLS element used for the Mass-Conserving mixed Stress (MCS)
    formulation for Stokes flow with weakly imposed stress symmetry.

    GLS^2(k) is the space of trace-free polynomials of degree k with
    continuous normal-tangential components.

    Reference: https://doi.org/10.1137/19M1248960

    Notes
    -----
    This element does not include the bubbles required for inf-sup stability of
    the weak symmetry constraint.

    """
    def __init__(self, ref_el, degree, variant=None):

        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        assert variant == "integral"

        if splitting is not None:
            ref_el = splitting(ref_el)

        poly_set = polynomial_set.TracelessTensorPolynomialSet(ref_el, degree)
        dual = GLSDual(ref_el, degree)
        sd = ref_el.get_spatial_dimension()
        formdegree = (1, sd-1)
        mapping = "covariant contravariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)


def GopalakrishnanLedererSchoberlFirstKind(ref_el, degree, variant=None):
    """The GLS element used for the Mass-Conserving mixed Stress (MCS)
    formulation for Stokes flow.

    GLS^1(k) is the space of trace-free polynomials of degree k with
    continuous normal-tangential components of degree k-1.

    Reference: https://doi.org/10.1093/imanum/drz022
    """
    fe = GopalakrishnanLedererSchoberlSecondKind(ref_el, degree, variant=variant)
    entity_dofs = fe.entity_dofs()
    sd = ref_el.get_spatial_dimension()
    facet = ref_el.construct_subelement(sd-1)
    dimPkm1 = (sd-1)*expansions.polynomial_dimension(facet, degree-1)

    indices = []
    for f in sorted(entity_dofs[sd-1]):
        indices.extend(entity_dofs[sd-1][f][:dimPkm1])
    for cell in sorted(entity_dofs[sd]):
        indices.extend(entity_dofs[sd][cell])

    return RestrictedElement(fe, indices=indices)



import numpy

from finat.fiat_elements import FiatElement
from finat.physically_mapped import identity, PhysicallyMappedElement
from gem import Literal, ListTensor
from copy import deepcopy


def determinant(A):
    """Return the determinant of A"""
    n = A.shape[0]
    if n == 0:
        return 1
    elif n == 1:
        return A[0, 0]
    elif n == 2:
        return A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    else:
        detA = A[0, 0] * determinant(A[1:, 1:])
        cols = numpy.ones(A.shape[1], dtype=bool)
        for j in range(1, n):
            cols[j] = False
            detA += (-1)**j * A[0, j] * determinant(A[1:][:, cols])
            cols[j] = True
        return detA


def adjugate(A):
    """Return the adjugate matrix of A"""
    A = numpy.asarray(A)
    C = numpy.zeros_like(A)
    rows = numpy.ones(A.shape[0], dtype=bool)
    cols = numpy.ones(A.shape[1], dtype=bool)
    for i in range(A.shape[0]):
        rows[i] = False
        for j in range(A.shape[1]):
            cols[j] = False
            C[j, i] = (-1)**(i+j)*determinant(A[rows, :][:, cols])
            cols[j] = True
        rows[i] = True
    return C


def piola_inverse(fiat_cell, J, detJ):
    """Return the basis transformation of evaluation at a point.
    This simply inverts the Piola transform inv(J / detJ) = adj(J)."""
    sd = fiat_cell.get_spatial_dimension()
    Jnp = numpy.array([[J[i, j] for j in range(sd)] for i in range(sd)])
    return adjugate(Jnp)


def normal_tangential_edge_transform(fiat_cell, J, detJ, f):
    """Return the basis transformation of
    normal and tangential edge moments"""
    R = numpy.array([[0, 1], [-1, 0]])
    that = fiat_cell.compute_edge_tangent(f)
    that /= numpy.linalg.norm(that)
    nhat = R @ that
    Jn = J @ Literal(nhat)
    Jt = J @ Literal(that)
    alpha = Jn @ Jt
    beta = Jt @ Jt
    # Compute the last row of inv([[1, 0], [alpha/detJ, beta/detJ]])
    row = (-1 * alpha / beta, detJ / beta)
    return row


def normal_tangential_face_transform(fiat_cell, J, detJ, f):
    """Return the basis transformation of
    normal and tangential face moments"""
    # Compute the reciprocal basis
    thats = fiat_cell.compute_tangents(2, f)
    nhat = numpy.cross(*thats)
    nhat /= numpy.dot(nhat, nhat)
    orth_vecs = numpy.array([nhat,
                             numpy.cross(nhat, thats[1]),
                             numpy.cross(thats[0], nhat)])
    # Compute A = (alpha, beta, gamma)
    Jts = J @ Literal(thats.T)
    Jorths = J @ Literal(orth_vecs.T)
    A = Jorths.T @ Jts
    # Compute the last two rows of inv([[1, 0, 0], A.T/detJ])
    det0 = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    det1 = A[2, 0] * A[0, 1] - A[2, 1] * A[0, 0]
    det2 = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    scale = detJ / det0
    rows = ((-1 * det1 / det0, -1 * scale * A[2, 1], scale * A[2, 0]),
            (-1 * det2 / det0, scale * A[1, 1], -1 * scale * A[1, 0]))
    return rows


class PiolaBubbleElement(PhysicallyMappedElement, FiatElement):
    """A general class to transform Piola-mapped elements with normal facet bubbles."""
    def __init__(self, fiat_element):
        mapping, = set(fiat_element.mapping())
        if mapping != "contravariant piola":
            raise ValueError(f"{type(fiat_element).__name__} needs to be Piola mapped.")
        super().__init__(fiat_element)

        # On each facet we expect the normal dof followed by the tangential ones
        # The tangential dofs should be numbered last, and are constrained to be zero
        sd = self.cell.get_spatial_dimension()
        reduced_dofs = deepcopy(self._element.entity_dofs())
        reduced_dim = 0
        cur = reduced_dofs[sd-1][0][0]
        for entity in sorted(reduced_dofs[sd-1]):
            reduced_dim += len(reduced_dofs[sd-1][entity][1:])
            reduced_dofs[sd-1][entity] = [cur]
            cur += 1
        self._entity_dofs = reduced_dofs
        self._space_dimension = fiat_element.space_dimension() - reduced_dim

    def entity_dofs(self):
        return self._entity_dofs

    def space_dimension(self):
        return self._space_dimension

    def basis_transformation(self, coordinate_mapping):
        sd = self.cell.get_spatial_dimension()
        bary, = self.cell.make_points(sd, 0, sd+1)
        J = coordinate_mapping.jacobian_at(bary)
        detJ = coordinate_mapping.detJ_at(bary)

        dofs = self.entity_dofs()
        bfs = self._element.entity_dofs()
        ndof = self.space_dimension()
        numbf = self._element.space_dimension()
        V = identity(numbf, ndof)

        # Undo the Piola transform for non-facet bubble basis functions
        nodes = self._element.get_dual_set().nodes
        Finv = piola_inverse(self.cell, J, detJ)
        for dim in dofs:
            if dim == sd-1:
                continue
            for e in sorted(dofs[dim]):
                k = 0
                while k < len(dofs[dim][e]):
                    cur = dofs[dim][e][k]
                    if len(nodes[cur].deriv_dict) > 0:
                        V[cur, cur] = detJ
                        k += 1
                    else:
                        s = dofs[dim][e][k:k+sd]
                        V[numpy.ix_(s, s)] = Finv
                        k += sd
        # Unpick the normal component for the facet bubbles
        if sd == 2:
            transform = normal_tangential_edge_transform
        elif sd == 3:
            transform = normal_tangential_face_transform

        for f in sorted(dofs[sd-1]):
            rows = numpy.asarray(transform(self.cell, J, detJ, f))
            cur_dofs = dofs[sd-1][f]
            cur_bfs = bfs[sd-1][f][1:]
            V[numpy.ix_(cur_bfs, cur_dofs)] = rows[..., :len(cur_dofs)]
        return ListTensor(V.T)



# This file was modified from FFC
# (http://bitbucket.org/fenics-project/ffc), copyright notice
# reproduced below.
#
# Copyright (C) 2009-2013 Kristian B. Oelgaard and Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import weakref
from functools import singledispatch, cache

import finat
import finat.ufl
import ufl

from FIAT import ufc_cell

__all__ = ("as_fiat_cell", "create_base_element",
           "create_element", "supported_elements")


# List of supported elements and mapping to element classes
supported_elements = {"Argyris": finat.Argyris,
                      "Bell": finat.Bell,
                      "Bernardi-Raugel": finat.BernardiRaugel,
                      "Bernardi-Raugel Bubble": finat.BernardiRaugelBubble,
                      "Bernstein": finat.Bernstein,
                      "Brezzi-Douglas-Fortin-Marini": finat.BrezziDouglasFortinMarini,
                      "Brezzi-Douglas-Marini Cube Face": finat.BrezziDouglasMariniCubeFace,
                      "Brezzi-Douglas-Marini": finat.BrezziDouglasMarini,
                      "Brezzi-Douglas-Marini Cube Edge": finat.BrezziDouglasMariniCubeEdge,
                      "Bubble": finat.Bubble,
                      "FacetBubble": finat.FacetBubble,
                      "Crouzeix-Raviart": finat.CrouzeixRaviart,
                      "Direct Serendipity": finat.DirectSerendipity,
                      "Discontinuous Lagrange": finat.DiscontinuousLagrange,
                      "Discontinuous Lagrange L2": finat.DiscontinuousLagrange,
                      "Discontinuous Taylor": finat.DiscontinuousTaylor,
                      "Discontinuous Raviart-Thomas": lambda *args, **kwargs: finat.DiscontinuousElement(finat.RaviartThomas(*args, **kwargs)),
                      "DPC": finat.DPC,
                      "DPC L2": finat.DPC,
                      "Hermite": finat.Hermite,
                      "Hsieh-Clough-Tocher": finat.HsiehCloughTocher,
                      "Reduced-Hsieh-Clough-Tocher": finat.ReducedHsiehCloughTocher,
                      "QuadraticPowellSabin6": finat.QuadraticPowellSabin6,
                      "QuadraticPowellSabin12": finat.QuadraticPowellSabin12,
                      "Alfeld-Sorokina": finat.AlfeldSorokina,
                      "Arnold-Qin": finat.ArnoldQin,
                      "Reduced-Arnold-Qin": finat.ReducedArnoldQin,
                      "Christiansen-Hu": finat.ChristiansenHu,
                      "Guzman-Neilan 1st kind H1": finat.GuzmanNeilanFirstKindH1,
                      "Guzman-Neilan 2nd kind H1": finat.GuzmanNeilanSecondKindH1,
                      "Guzman-Neilan H1(div)": finat.GuzmanNeilanH1div,
                      "Guzman-Neilan Bubble": finat.GuzmanNeilanBubble,
                      "Johnson-Mercier": finat.JohnsonMercier,
                      "Lagrange": finat.Lagrange,
                      "Kong-Mulder-Veldhuizen": finat.KongMulderVeldhuizen,
                      "Gauss-Lobatto-Legendre": finat.GaussLobattoLegendre,
                      "Gauss-Legendre": finat.GaussLegendre,
                      "Gauss-Legendre L2": finat.GaussLegendre,
                      "Morley": finat.Morley,
                      "Nedelec 1st kind H(curl)": finat.Nedelec,
                      "Nedelec 2nd kind H(curl)": finat.NedelecSecondKind,
                      "Raviart-Thomas": finat.RaviartThomas,
                      "Real": finat.Real,
                      "S": finat.Serendipity,
                      "SminusF": finat.TrimmedSerendipityFace,
                      "SminusDiv": finat.TrimmedSerendipityDiv,
                      "SminusE": finat.TrimmedSerendipityEdge,
                      "SminusCurl": finat.TrimmedSerendipityCurl,
                      "Regge": finat.Regge,
                      "HDiv Trace": finat.HDivTrace,
                      "Hellan-Herrmann-Johnson": finat.HellanHerrmannJohnson,
                      "Gopalakrishnan-Lederer-Schoberl 1st kind": finat.GopalakrishnanLedererSchoberlFirstKind,
                      "Gopalakrishnan-Lederer-Schoberl 2nd kind": finat.GopalakrishnanLedererSchoberlSecondKind,
                      "Conforming Arnold-Winther": finat.ArnoldWinther,
                      "Nonconforming Arnold-Winther": finat.ArnoldWintherNC,
                      "Hu-Zhang": finat.HuZhang,
                      "Mardal-Tai-Winther": finat.MardalTaiWinther,
                      # These require special treatment
                      "Q": None,
                      "DQ": None,
                      "DQ L2": None,
                      "RTCE": None,
                      "RTCF": None,
                      "NCE": None,
                      "NCF": None,
                      }
"""A :class:`.dict` mapping UFL element family names to their
FInAT-equivalent constructors.  If the value is ``None``, the UFL
element is supported, but must be handled specially because it doesn't
have a direct FInAT equivalent."""


@cache
def as_fiat_cell(cell):
    """Convert a ufl cell to a FIAT cell.

    :arg cell: the :class:`ufl.Cell` to convert."""
    if not isinstance(cell, ufl.AbstractCell):
        raise ValueError("Expecting a UFL Cell")
    return ufc_cell(cell)


@singledispatch
def convert(element, **kwargs):
    """Handler for converting UFL elements to FInAT elements.

    :arg element: The UFL element to convert.

    Do not use this function directly, instead call
    :func:`create_element`."""
    if element.family() in supported_elements:
        raise ValueError("Element %s supported, but no handler provided" % element)
    raise ValueError("Unsupported element type %s" % type(element))


cg_interval_variants = {
    "fdm": finat.FDMLagrange,
    "fdm_ipdg": finat.FDMLagrange,
    "fdm_quadrature": finat.FDMQuadrature,
    "fdm_broken": finat.FDMBrokenH1,
    "fdm_hermite": finat.FDMHermite,
}


dg_interval_variants = {
    "fdm": finat.FDMDiscontinuousLagrange,
    "fdm_quadrature": finat.FDMDiscontinuousLagrange,
    "fdm_ipdg": lambda *args: finat.DiscontinuousElement(finat.FDMLagrange(*args)),
    "fdm_broken": finat.FDMBrokenL2,
}


# Base finite elements first
@convert.register(finat.ufl.FiniteElement)
def convert_finiteelement(element, **kwargs):
    cell = as_fiat_cell(element.cell)
    if element.family() in {"Quadrature", "Boundary Quadrature"}:
        degree = element.degree()
        scheme = element.quadrature_scheme() or "default"
        if degree is None or scheme is None:
            raise ValueError("Quadrature scheme and degree must be specified!")

        codim = 1 if element.family() == "Boundary Quadrature" else 0
        return finat.make_quadrature_element(cell, degree, scheme, codim), set()
    lmbda = supported_elements[element.family()]
    if element.family() == "Real" and element.cell.cellname() in {"quadrilateral", "hexahedron"}:
        lmbda = None
        element = finat.ufl.FiniteElement("DQ", element.cell, 0)
    if lmbda is None:
        if element.cell.cellname() == "quadrilateral":
            # Handle quadrilateral short names like RTCF and RTCE.
            element = element.reconstruct(cell=quadrilateral_tpc)
        elif element.cell.cellname() == "hexahedron":
            # Handle hexahedron short names like NCF and NCE.
            element = element.reconstruct(cell=hexahedron_tpc)
        else:
            raise ValueError("%s is supported, but handled incorrectly" %
                             element.family())
        finat_elem, deps = _create_element(element, **kwargs)
        return finat.FlattenedDimensions(finat_elem), deps

    finat_kwargs = {}
    kind = element.variant()
    if kind is None:
        kind = 'spectral'  # default variant

    if element.family() == "Lagrange":
        if kind in ['spectral', 'mimetic']:
            lmbda = finat.GaussLobattoLegendre
        elif element.cell.cellname() == "interval" and kind in cg_interval_variants:
            lmbda = cg_interval_variants[kind]
        elif any(map(kind.startswith, ['integral', 'demkowicz', 'fdm'])):
            lmbda = finat.IntegratedLegendre
            finat_kwargs["variant"] = kind
        elif kind in ['mgd', 'feec', 'qb', 'mse']:
            degree = element.degree()
            shift_axes = kwargs["shift_axes"]
            restriction = kwargs["restriction"]
            deps = {"shift_axes", "restriction"}
            return finat.RuntimeTabulated(cell, degree, variant=kind, shift_axes=shift_axes, restriction=restriction), deps
        else:
            # Let FIAT handle the general case
            lmbda = finat.Lagrange
            finat_kwargs["variant"] = kind

    elif element.family() in ["Discontinuous Lagrange", "Discontinuous Lagrange L2"]:
        if kind == 'spectral':
            lmbda = finat.GaussLegendre
        elif kind == 'mimetic':
            lmbda = finat.Histopolation
        elif element.cell.cellname() == "interval" and kind in dg_interval_variants:
            lmbda = dg_interval_variants[kind]
        elif any(map(kind.startswith, ['integral', 'demkowicz', 'fdm'])):
            lmbda = finat.Legendre
            finat_kwargs["variant"] = kind
        elif kind in ['mgd', 'feec', 'qb', 'mse']:
            degree = element.degree()
            shift_axes = kwargs["shift_axes"]
            restriction = kwargs["restriction"]
            deps = {"shift_axes", "restriction"}
            return finat.RuntimeTabulated(cell, degree, variant=kind, shift_axes=shift_axes, restriction=restriction, continuous=False), deps
        else:
            # Let FIAT handle the general case
            lmbda = finat.DiscontinuousLagrange
            finat_kwargs["variant"] = kind

    elif element.family() in {"HDiv Trace", "Bubble", "FacetBubble"}:
        finat_kwargs["variant"] = kind

    elif element.variant() is not None:
        finat_kwargs["variant"] = element.variant()

    return lmbda(cell, element.degree(), **finat_kwargs), set()


# Element modifiers and compound element types
@convert.register(finat.ufl.BrokenElement)
def convert_brokenelement(element, **kwargs):
    finat_elem, deps = _create_element(element._element, **kwargs)
    return finat.DiscontinuousElement(finat_elem), deps


@convert.register(finat.ufl.EnrichedElement)
def convert_enrichedelement(element, **kwargs):
    elements, deps = zip(*[_create_element(elem, **kwargs)
                           for elem in element._elements])
    return finat.EnrichedElement(elements), set.union(*deps)


@convert.register(finat.ufl.NodalEnrichedElement)
def convert_nodalenrichedelement(element, **kwargs):
    elements, deps = zip(*[_create_element(elem, **kwargs)
                           for elem in element._elements])
    return finat.NodalEnrichedElement(elements), set.union(*deps)


@convert.register(finat.ufl.MixedElement)
def convert_mixedelement(element, **kwargs):
    elements, deps = zip(*[_create_element(elem, **kwargs)
                           for elem in element.sub_elements])
    return finat.MixedElement(elements), set.union(*deps)


@convert.register(finat.ufl.VectorElement)
@convert.register(finat.ufl.TensorElement)
def convert_tensorelement(element, **kwargs):
    inner_elem, deps = _create_element(element.sub_elements[0], **kwargs)
    shape = element.reference_value_shape
    shape = shape[:len(shape) - len(inner_elem.value_shape)]
    shape_innermost = kwargs["shape_innermost"]
    return (finat.TensorFiniteElement(inner_elem, shape, not shape_innermost),
            deps | {"shape_innermost"})


@convert.register(finat.ufl.TensorProductElement)
def convert_tensorproductelement(element, **kwargs):
    cell = element.cell
    if type(cell) is not ufl.TensorProductCell:
        raise ValueError("TensorProductElement not on TensorProductCell?")
    shift_axes = kwargs["shift_axes"]
    dim_offset = 0
    elements = []
    deps = set()
    for elem in element.sub_elements:
        kwargs["shift_axes"] = shift_axes + dim_offset
        dim_offset += elem.cell.topological_dimension()
        finat_elem, ds = _create_element(elem, **kwargs)
        elements.append(finat_elem)
        deps.update(ds)
    return finat.TensorProductElement(elements), deps


@convert.register(finat.ufl.HDivElement)
def convert_hdivelement(element, **kwargs):
    finat_elem, deps = _create_element(element._element, **kwargs)
    return finat.HDivElement(finat_elem), deps


@convert.register(finat.ufl.HCurlElement)
def convert_hcurlelement(element, **kwargs):
    finat_elem, deps = _create_element(element._element, **kwargs)
    return finat.HCurlElement(finat_elem), deps


@convert.register(finat.ufl.WithMapping)
def convert_withmapping(element, **kwargs):
    return _create_element(element.wrapee, **kwargs)


@convert.register(finat.ufl.RestrictedElement)
def convert_restrictedelement(element, **kwargs):
    finat_elem, deps = _create_element(element._element, **kwargs)
    return finat.RestrictedElement(finat_elem, element.restriction_domain()), deps


hexahedron_tpc = ufl.TensorProductCell(ufl.interval, ufl.interval, ufl.interval)
quadrilateral_tpc = ufl.TensorProductCell(ufl.interval, ufl.interval)
_cache = weakref.WeakKeyDictionary()


def create_element(ufl_element, shape_innermost=True, shift_axes=0, restriction=None):
    """Create a FInAT element (suitable for tabulating with) given a UFL element.

    :arg ufl_element: The UFL element to create a FInAT element from.
    :arg shape_innermost: Vector/tensor indices come after basis function indices
    :arg restriction: cell restriction in interior facet integrals
                      (only for runtime tabulated elements)
    """
    finat_element, deps = _create_element(ufl_element,
                                          shape_innermost=shape_innermost,
                                          shift_axes=shift_axes,
                                          restriction=restriction)
    return finat_element


def _create_element(ufl_element, **kwargs):
    """A caching wrapper around :py:func:`convert`.

    Takes a UFL element and an unspecified set of parameter options,
    and returns the converted element with the set of keyword names
    that were relevant for conversion.
    """
    # Look up conversion in cache
    try:
        cache = _cache[ufl_element]
    except KeyError:
        _cache[ufl_element] = {}
        cache = _cache[ufl_element]

    for key, finat_element in cache.items():
        # Cache hit if all relevant parameter values match.
        if all(kwargs[param] == value for param, value in key):
            return finat_element, set(param for param, value in key)

    # Convert if cache miss
    if ufl_element.cell is None:
        raise ValueError("Don't know how to build element when cell is not given")

    finat_element, deps = convert(ufl_element, **kwargs)

    # Store conversion in cache
    key = frozenset((param, kwargs[param]) for param in deps)
    cache[key] = finat_element

    # Forward result
    return finat_element, deps


def create_base_element(ufl_element, **kwargs):
    """Create a "scalar" base FInAT element given a UFL element.
    Takes a UFL element and an unspecified set of parameter options,
    and returns the converted element.
    """
    finat_element = create_element(ufl_element, **kwargs)
    if isinstance(finat_element, finat.TensorFiniteElement):
        finat_element = finat_element.base_element
    return finat_element



from FIAT.reference_element import (
    UFCQuadrilateral, UFCInterval, TensorProductCell)
from FIAT import Serendipity
import numpy as np
import sympy


def test_serendipity_derivatives():
    cell = UFCQuadrilateral()
    S = Serendipity(cell, 2)

    x = sympy.DeferredVector("X")
    X, Y = x[0], x[1]
    basis_functions = [
        (1 - X)*(1 - Y),
        Y*(1 - X),
        X*(1 - Y),
        X*Y,
        Y*(1 - X)*(Y - 1),
        X*Y*(Y - 1),
        X*(1 - Y)*(X - 1),
        X*Y*(X - 1),
    ]
    points = [[0.5, 0.5], [0.25, 0.75]]
    for alpha, actual in S.tabulate(2, points).items():
        expect = list(sympy.diff(basis, *zip([X, Y], alpha))
                      for basis in basis_functions)
        expect = list([basis.subs(dict(zip([X, Y], point)))
                       for point in points]
                      for basis in expect)
        assert actual.shape == (8, 2)
        assert np.allclose(np.asarray(expect, dtype=float),
                           actual.reshape(8, 2))


def test_dual_tensor_versus_ufc():
    K0 = UFCQuadrilateral()
    ell = UFCInterval()
    K1 = TensorProductCell(ell, ell)
    S0 = Serendipity(K0, 2)
    S1 = Serendipity(K1, 2)
    # since both elements go through the flattened cell to produce the
    # dual basis, they ought to do *exactly* the same calculations and
    # hence form exactly the same nodes.
    for i in range(S0.space_dimension()):
        assert S0.dual.nodes[i].pt_dict == S1.dual.nodes[i].pt_dict



# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

from FIAT import finite_element, polynomial_set
from FIAT.bernardi_raugel import BernardiRaugelDualSet
from FIAT.quadrature_schemes import create_quadrature
from FIAT.reference_element import TRIANGLE
from FIAT.macro import CkPolynomialSet
from FIAT.hct import HsiehCloughTocher

import numpy


def ArnoldQinSpace(ref_el, degree, reduced=False):
    """Return a basis for the Arnold-Qin space
    curl(HCT-red) + P0 x if reduced = True, and
    curl(HCT) + P0 x if reduced = False."""
    if ref_el.get_shape() != TRIANGLE:
        raise ValueError("Arnold-Qin only defined on triangles")
    if degree != 2:
        raise ValueError("Arnold-Qin only defined for degree = 2")
    sd = ref_el.get_spatial_dimension()
    HCT = HsiehCloughTocher(ref_el, degree+1, reduced=True)
    ref_complex = HCT.get_reference_complex()
    Q = create_quadrature(ref_complex, 2 * degree)
    qpts, qwts = Q.get_points(), Q.get_weights()

    x = qpts.T
    bary = numpy.asarray(ref_el.make_points(sd, 0, sd+1))
    P0x_at_qpts = x[None, :, :] - bary[:, :, None]

    tab = HCT.tabulate(1, qpts)
    curl_at_qpts = numpy.stack([tab[(0, 1)], -tab[(1, 0)]], axis=1)
    if reduced:
        curl_at_qpts = curl_at_qpts[:9]

    C0 = CkPolynomialSet(ref_complex, degree, order=0, scale=1, variant="bubble")
    C0_at_qpts = C0.tabulate(qpts)[(0,) * sd]
    duals = numpy.multiply(C0_at_qpts, qwts)
    M = numpy.dot(duals, C0_at_qpts.T)
    duals = numpy.linalg.solve(M, duals)

    # Remove the constant nullspace
    ids = [0, 3, 6]
    A = numpy.asarray([[1, 1, 1], [1, -1, 0], [0, -1, 1]])
    phis = curl_at_qpts
    phis[ids] = numpy.tensordot(A, phis[ids], axes=(-1, 0))
    # Replace the constant nullspace with P_0 x
    phis[0] = P0x_at_qpts
    coeffs = numpy.tensordot(phis, duals, axes=(-1, -1))
    return polynomial_set.PolynomialSet(ref_complex, degree, degree,
                                        C0.get_expansion_set(), coeffs)


class ArnoldQin(finite_element.CiarletElement):
    """The Arnold-Qin C0(Alfeld) quadratic macroelement with divergence in P0.
    This element belongs to a Stokes complex, and is paired with unsplit DG0."""
    def __init__(self, ref_el, degree=2, reduced=False):
        poly_set = ArnoldQinSpace(ref_el, degree)
        if reduced:
            order = 1
            mapping = "contravariant piola"
        else:
            order = degree
            mapping = "affine"
        dual = BernardiRaugelDualSet(ref_el, order, degree=degree)
        formdegree = ref_el.get_spatial_dimension() - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



import pytest
import FIAT
import gem
import numpy as np
from finat.physically_mapped import PhysicalGeometry


class MyMapping(PhysicalGeometry):
    def __init__(self, ref_cell, phys_cell):
        self.ref_cell = ref_cell
        self.phys_cell = phys_cell

        self.A, self.b = FIAT.reference_element.make_affine_mapping(
            self.ref_cell.vertices,
            self.phys_cell.vertices)

    def cell_size(self):
        # Currently, just return 1 so we can compare FIAT dofs
        # to transformed dofs.
        return np.ones((len(self.ref_cell.vertices),))

    def detJ_at(self, point):
        return gem.Literal(np.linalg.det(self.A))

    def jacobian_at(self, point):
        return gem.Literal(self.A)

    def normalized_reference_edge_tangents(self):
        top = self.ref_cell.get_topology()
        return gem.Literal(
            np.asarray([self.ref_cell.compute_normalized_edge_tangent(i)
                        for i in sorted(top[1])]))

    def reference_normals(self):
        sd = self.ref_cell.get_spatial_dimension()
        top = self.ref_cell.get_topology()
        return gem.Literal(
            np.asarray([self.ref_cell.compute_normal(i)
                        for i in sorted(top[sd-1])]))

    def physical_normals(self):
        sd = self.phys_cell.get_spatial_dimension()
        top = self.phys_cell.get_topology()
        return gem.Literal(
            np.asarray([self.phys_cell.compute_normal(i)
                        for i in sorted(top[sd-1])]))

    def physical_tangents(self):
        top = self.phys_cell.get_topology()
        return gem.Literal(
            np.asarray([self.phys_cell.compute_normalized_edge_tangent(i)
                        for i in sorted(top[1])]))

    def physical_edge_lengths(self):
        top = self.phys_cell.get_topology()
        return gem.Literal(
            np.asarray([self.phys_cell.volume_of_subcomplex(1, i)
                        for i in sorted(top[1])]))

    def physical_points(self, ps, entity=None):
        prefs = ps.points
        A, b = self.A, self.b
        return gem.Literal(np.asarray([A @ x + b for x in prefs]))

    def physical_vertices(self):
        return gem.Literal(self.phys_cell.verts)


class ScaledMapping(MyMapping):

    def cell_size(self):
        # Firedrake interprets this as 2x the circumradius
        sd = self.phys_cell.get_spatial_dimension()
        top = self.phys_cell.get_topology()
        vol = self.phys_cell.volume()
        edges = [self.phys_cell.volume_of_subcomplex(1, i)
                 for i in sorted(top[1])]

        if sd == 1:
            cs = vol
        elif sd == 2:
            cs = np.prod(edges) / (2 * vol)
        elif sd == 3:
            edge_pairs = [edges[i] * edges[j] for i in top[1] for j in top[1]
                          if len(set(top[1][i] + top[1][j])) == len(top[0])]
            cs = 1.0 / (12 * vol)
            for k in range(4):
                s = [1]*len(edge_pairs)
                if k > 0:
                    s[k-1] = -1
                cs *= np.dot(s, edge_pairs) ** 0.5
        else:
            raise NotImplementedError(f"Cell size not implemented in {sd} dimensions")
        return np.asarray([cs for _ in sorted(top[0])])


def scaled_simplex(dim, scale):
    K = FIAT.ufc_simplex(dim)
    K.vertices = scale * np.array(K.vertices)
    return K


@pytest.fixture
def ref_el():
    K = {dim: FIAT.ufc_simplex(dim) for dim in (2, 3)}
    return K


@pytest.fixture
def phys_el():
    K = {dim: FIAT.ufc_simplex(dim) for dim in (2, 3)}
    K[2].vertices = ((0.0, 0.1), (1.17, -0.09), (0.15, 1.84))
    K[3].vertices = ((0, 0, 0),
                     (1., 0.1, -0.37),
                     (0.01, 0.987, -.23),
                     (-0.1, -0.2, 1.38))
    return K


@pytest.fixture
def ref_to_phys(ref_el, phys_el):
    return {dim: MyMapping(ref_el[dim], phys_el[dim]) for dim in ref_el}


@pytest.fixture
def scaled_ref_to_phys(ref_el):
    return {dim: [ScaledMapping(ref_el[dim], scaled_simplex(dim, 0.5**k)) for k in range(3)]
            for dim in ref_el}



from functools import singledispatch
from itertools import chain

import FIAT
from FIAT.polynomial_set import mis

import finat
from finat.fiat_elements import FiatElement
from finat.physically_mapped import PhysicallyMappedElement


# Sentinel for when restricted element is empty
null_element = object()


@singledispatch
def restrict(element, domain, take_closure):
    """Restrict an element to a given subentity.

    :arg element: The element to restrict.
    :arg domain: The subentity to restrict to.
    :arg take_closure: Gather dofs in closure of the subentities?
        Ignored for "interior" domain.

    :raises NotImplementedError: If we don't know how to restrict this
        element.
    :raises ValueError: If the restricted element is empty.
    :returns: A new finat element."""
    return NotImplementedError(f"Don't know how to restrict element of type {type(element)}")


@restrict.register(FiatElement)
def restrict_fiat(element, domain, take_closure):
    try:
        return FiatElement(FIAT.RestrictedElement(element._element,
                           restriction_domain=domain, take_closure=take_closure))
    except ValueError:
        return null_element


@restrict.register(PhysicallyMappedElement)
def restrict_physically_mapped(element, domain, take_closure):
    raise NotImplementedError("Can't restrict Physically Mapped things")


@restrict.register(finat.FlattenedDimensions)
def restrict_flattened_dimensions(element, domain, take_closure):
    restricted = restrict(element.product, domain, take_closure)
    if restricted is null_element:
        return null_element
    else:
        return finat.FlattenedDimensions(restricted)


@restrict.register(finat.DiscontinuousElement)
@restrict.register(finat.DiscontinuousLagrange)
@restrict.register(finat.Legendre)
def restrict_discontinuous(element, domain, take_closure):
    if domain == "interior":
        return element
    else:
        return null_element


@restrict.register(finat.EnrichedElement)
def restrict_enriched(element, domain, take_closure):
    if all(isinstance(e, finat.mixed.MixedSubElement) for e in element.elements):
        # Mixed is handled by Enriched + MixedSubElement, we must
        # restrict the subelements here because the transformation is
        # nonlocal.
        elements = tuple(restrict(e.element, domain, take_closure) for
                         e in element.elements)
        reconstruct = finat.mixed.MixedElement
    elif not any(isinstance(e, finat.mixed.MixedSubElement) for e in element.elements):
        elements = tuple(restrict(e, domain, take_closure)
                         for e in element.elements)
        reconstruct = finat.EnrichedElement
    else:
        raise NotImplementedError("Not expecting enriched with mixture of MixedSubElement and others")

    elements = tuple(e for e in elements if e is not null_element)
    if elements:
        return reconstruct(elements)
    else:
        return null_element


@restrict.register(finat.HCurlElement)
def restrict_hcurl(element, domain, take_closure):
    restricted = restrict(element.wrappee, domain, take_closure)
    if restricted is null_element:
        return null_element
    else:
        if isinstance(restricted, finat.EnrichedElement):
            return finat.EnrichedElement(finat.HCurlElement(e)
                                         for e in restricted.elements)
        else:
            return finat.HCurlElement(restricted)


@restrict.register(finat.HDivElement)
def restrict_hdiv(element, domain, take_closure):
    restricted = restrict(element.wrappee, domain, take_closure)
    if restricted is null_element:
        return null_element
    else:
        if isinstance(restricted, finat.EnrichedElement):
            return finat.EnrichedElement(finat.HDivElement(e)
                                         for e in restricted.elements)
        else:
            return finat.HDivElement(restricted)


@restrict.register(finat.mixed.MixedSubElement)
def restrict_mixed(element, domain, take_closure):
    raise AssertionError("Was expecting this to be handled inside EnrichedElement restriction")


def r_to_codim(restriction, dim):
    if restriction == "interior":
        return 0
    elif restriction == "facet":
        return 1
    elif restriction == "face":
        return dim - 2
    elif restriction == "edge":
        return dim - 1
    elif restriction == "vertex":
        return dim
    else:
        raise ValueError


def codim_to_r(codim, dim):
    d = dim - codim
    if codim == 0:
        return "interior"
    elif codim == 1:
        return "facet"
    elif d == 0:
        return "vertex"
    elif d == 1:
        return "edge"
    elif d == 2:
        return "face"
    else:
        raise ValueError


@restrict.register(finat.TensorProductElement)
def restrict_tpe(element, domain, take_closure):
    # The restriction of a TPE to a codim subentity is the direct sum
    # of TPEs where the factors have been restricted in such a way
    # that the sum of those restrictions is codim.
    #
    # For example, to restrict an interval x interval to edges (codim 1)
    # we construct
    #
    # R(I, 0)⊗R(I, 1) ⊕ R(I, 1)⊗R(I, 0)
    #
    # If take_closure is true, the restriction wants to select dofs on
    # entities with dim >= codim >= 1 (for the edge example)
    # so we get
    #
    # R(I, 0)⊗R(I, 1) ⊕ R(I, 1)⊗R(I, 0) ⊕ R(I, 0)⊗R(I, 0)
    factors = element.factors
    dimension = element.cell.get_spatial_dimension()
    # Figure out which codim entity we're selecting
    codim = r_to_codim(domain, dimension)
    # And the range of codims.
    upper = 1 + (dimension
                 if (take_closure and domain != "interior")
                 else codim)
    # restrictions on each factor taken from n-tuple that sums to the
    # target codim (as long as the codim <= dim_factor)
    restrictions = tuple(candidate
                         for candidate in chain(*(mis(len(factors), c)
                                                  for c in range(codim, upper)))
                         if all(d <= factor.cell.get_dimension()
                                for d, factor in zip(candidate, factors)))
    take_closure = False
    elements = []
    for decomposition in restrictions:
        # Recurse, but don't take closure in recursion (since we
        # handled it already).
        new_factors = tuple(
            restrict(factor, codim_to_r(codim, factor.cell.get_dimension()),
                     take_closure)
            for factor, codim in zip(factors, decomposition))
        # If one of the factors was empty then the whole TPE is empty,
        # so skip.
        if all(f is not null_element for f in new_factors):
            elements.append(finat.TensorProductElement(new_factors))
    if elements:
        return finat.EnrichedElement(elements)
    else:
        return null_element


@restrict.register(finat.TensorFiniteElement)
def restrict_tfe(element, domain, take_closure):
    restricted = restrict(element._base_element, domain, take_closure)
    if restricted is null_element:
        return null_element
    else:
        return finat.TensorFiniteElement(restricted, element._shape, element._transpose)


@restrict.register(finat.HDivTrace)
def restrict_hdivtrace(element, domain, take_closure):
    try:
        return FiatElement(FIAT.RestrictedElement(element._element, restriction_domain=domain))
    except ValueError:
        return null_element


def RestrictedElement(element, restriction_domain, *, indices=None):
    """Construct a restricted element.

    :arg element: The element to restrict.
    :arg restriction_domain: Which entities to restrict to.
    :arg indices: Indices of basis functions to select (not supported)
    :returns: A new element.

    .. note::

       A restriction domain of "interior" means to select the dofs on
       the cell, all other domains (e.g. "face", "edge") select dofs
       in the closure of the entity.

    .. warning::

       The returned element *may not* be interpolatory. That is, the
       dual basis (if implemented) might not be nodal to the primal
       basis. Assembly still works (``basis_evaluation`` is fine), but
       interpolation may produce bad results.

       Restrictions of FIAT-implemented CiarletElements are always
       nodal.
    """
    if indices is not None:
        raise NotImplementedError("Only done for topological restrictions")
    assert restriction_domain is not None
    restricted = restrict(element, restriction_domain, take_closure=True)
    if restricted is null_element:
        raise ValueError("Restricted element is empty")
    return restricted



import FIAT

from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase


class DiscontinuousElement(FiniteElementBase):
    """Element wrapper that makes a FInAT element discontinuous."""

    def __init__(self, element):
        super().__init__()
        self.element = element

    @property
    def cell(self):
        return self.element.cell

    @property
    def complex(self):
        return self.element.complex

    @property
    def degree(self):
        return self.element.degree

    @cached_property
    def formdegree(self):
        # Always discontinuous!
        return self.element.cell.get_spatial_dimension()

    @cached_property
    def _entity_dofs(self):
        result = {dim: {i: [] for i in entities}
                  for dim, entities in self.cell.get_topology().items()}
        cell_dimension = self.cell.get_dimension()
        result[cell_dimension][0].extend(range(self.space_dimension()))
        return result

    def entity_dofs(self):
        return self._entity_dofs

    @cached_property
    def entity_permutations(self):
        # Return entity_permutations of the base finite element if it only
        # has cell degrees of freedom; otherwise entity_permutations is not
        # yet implemented for DiscontinuousElement.
        if self.element.entity_dofs() == self.element.entity_closure_dofs():
            return self.element.entity_permutations
        else:
            raise NotImplementedError(f"entity_permutations not yet implemented for a general {type(self)}")

    def space_dimension(self):
        return self.element.space_dimension()

    @property
    def index_shape(self):
        return self.element.index_shape

    @property
    def value_shape(self):
        return self.element.value_shape

    @cached_property
    def fiat_equivalent(self):
        return FIAT.DiscontinuousElement(self.element.fiat_equivalent)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        return self.element.basis_evaluation(order, ps, entity, coordinate_mapping=coordinate_mapping)

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        return self.element.point_evaluation(order, refcoords, entity, coordinate_mapping)

    @property
    def dual_basis(self):
        return self.element.dual_basis

    @property
    def mapping(self):
        return self.element.mapping



# Copyright (C) 2020 Robert C. Kirby (Baylor University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
import numpy as np
from FIAT.functional import Functional
from FIAT.dual_set import DualSet
from collections import defaultdict
from itertools import zip_longest


def compute_pointwise_dual(el, pts):
    """Constructs a dual basis to the basis for el as a linear combination
    of a set of pointwise evaluations.  This is useful when the
    prescribed finite element isn't Ciarlet (e.g. the basis functions
    are provided explicitly as formulae).  Alternately, the element's
    given dual basis may involve differentiation, making run-time
    interpolation difficult in FIAT clients.  The pointwise dual,
    consisting only of pointwise evaluations, will effectively replace
    these derivatives with (automatically determined) finite
    differences.  This is exact on the polynomial space, but is an
    approximation if applied to functions outside the space.

    :param el: a :class:`FiniteElement`.
    :param pts: an iterable of points with the same length as el's
                dimension.  These points must be unisolvent for the
                polynomial space
    :returns: a :class `DualSet`
    """
    nbf = el.space_dimension()

    T = el.ref_el
    sd = T.get_spatial_dimension()

    assert np.asarray(pts).shape == (int(nbf / np.prod(el.value_shape())), sd)

    z = tuple([0] * sd)

    nds = []

    V = el.tabulate(0, pts)[z]

    # Make a square system, invert, and then put it back in the right
    # shape so we have (nbf, ..., npts) with more dimensions
    # for vector or tensor-valued elements.
    alphas = np.linalg.inv(V.reshape((nbf, -1)).T).reshape(V.shape)

    # Each row of alphas gives the coefficients of a functional,
    # represented, as elsewhere in FIAT, as a summation of
    # components of the input at particular points.

    # This logic picks out the points and components for which the
    # weights are actually nonzero to construct the functional.

    pts = np.asarray(pts)
    for coeffs in alphas:
        pt_dict = defaultdict(list)
        nonzero = np.where(np.abs(coeffs) > 1.e-12)
        *comp, pt_index = nonzero

        for pt, coeff_comp in zip(pts[pt_index],
                                  zip_longest(coeffs[nonzero],
                                              zip(*comp), fillvalue=())):
            pt_dict[tuple(pt)].append(coeff_comp)

        nds.append(Functional(T, el.value_shape(), dict(pt_dict), {}, "node"))

    return DualSet(nds, T, el.entity_dofs())



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# Modified 2020 by the same from Baylor University
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# functionals require:
# - a degree of accuracy (-1 indicates that it works for all functions
#   such as point evaluation)
# - a reference element domain
# - type information

from itertools import chain
import numpy

from FIAT import polynomial_set, jacobi, quadrature_schemes


def index_iterator(shp):
    """Constructs a generator iterating over all indices in
    shp in generalized column-major order  So if shp = (2,2), then we
    construct the sequence (0,0),(0,1),(1,0),(1,1)"""
    return numpy.ndindex(shp)


class Functional(object):
    r"""Abstract class representing a linear functional.
    All FIAT functionals are discrete in the sense that
    they are written as a weighted sum of (derivatives of components of) their
    argument evaluated at particular points.

    :arg ref_el: a :class:`Cell`
    :arg target_shape: a tuple indicating the value shape of functions on
         the functional operates (e.g. if the function eats 2-vectors
         then target_shape is (2,) and if it eats scalars then
         target_shape is ()
    :arg pt_dict: A dict mapping points to lists of information about
         how the functional is evaluated.  Each entry in the list takes
         the form of a tuple (wt, comp) so that (at least if the
         deriv_dict argument is empty), the functional takes the form
         :math:`\ell(f) = \sum_{q=1}^{N_q} \sum_{k=1}^{K_q} w^q_k f_{c_k}(x_q)`
         where :math:`f_{c_k}` indicates a particular vector or tensor component
    :arg deriv_dict: A dict that is similar to `pt_dict`, although the entries
         of each list are tuples (wt, alpha, comp) with alpha a tuple
         of nonnegative integers corresponding to the order of partial
         differentiation in each spatial direction.
    :arg functional_type: a string labeling the kind of functional
         this is.
    """
    def __init__(self, ref_el, target_shape, pt_dict, deriv_dict,
                 functional_type):
        self.ref_el = ref_el
        self.target_shape = target_shape
        self.pt_dict = pt_dict
        self.deriv_dict = deriv_dict
        self.functional_type = functional_type
        if len(deriv_dict) > 0:
            self.max_deriv_order = max(sum(wac[1]) for wac in chain(*deriv_dict.values()))
        else:
            self.max_deriv_order = 0

    def evaluate(self, f):
        """Obsolete and broken functional evaluation.

        To evaluate the functional, call it on the target function:

          functional(function)
        """
        raise AttributeError("To evaluate the functional just call it on a function.")

    def __call__(self, fn):
        raise NotImplementedError("Evaluation is not yet implemented for %s" % type(self))

    def get_point_dict(self):
        """Returns the functional information, which is a dictionary
        mapping each point in the support of the functional to a list
        of pairs containing the weight and component."""
        return self.pt_dict

    def get_reference_element(self):
        """Returns the reference element."""
        return self.ref_el

    def get_type_tag(self):
        """Returns the type of function (e.g. point evaluation or
        normal component, which is probably handy for clients of FIAT"""
        return self.functional_type

    def to_riesz(self, poly_set):
        r"""Constructs an array representation of the functional so
        that the functional may be applied to a function expressed in
        in terms of the expansion set underlying  `poly_set` by means
        of contracting coefficients.

        That is, `poly_set` will have members all expressed in the
        form :math:`p = \sum_{i} \alpha^i \phi_i`
        where :math:`\{\phi_i\}_{i}` is some orthonormal expansion set
        and :math:`\alpha^i` are coefficients.  Note: the orthonormal
        expansion set is always scalar-valued but if the members of
        `poly_set` are vector or tensor valued the :math:`\alpha^i`
        will be scalars or vectors.

        This function constructs a tensor :math:`R` such that the
        contraction of :math:`R` with the array of coefficients
        :math:`\alpha` produces the effect of :math:`\ell(f)`

        In the case of scalar-value functions, :math:`R` is just a
        vector of the same length as the expansion set, and
        :math:`R_i = \ell(\phi_i)`.  For vector-valued spaces,
        :math:`R_{ij}` will be :math:`\ell(e^i \phi_j)` where
        :math:`e^i` is the canonical unit vector nonzero only in one
        entry :math:`i`.
        """
        es = poly_set.get_expansion_set()
        ed = poly_set.get_embedded_degree()
        nexp = es.get_num_members(ed)

        pt_dict = self.get_point_dict()

        pts = list(pt_dict.keys())
        npts = len(pts)

        bfs = es.tabulate(ed, pts)
        result = numpy.zeros(poly_set.coeffs.shape[1:], "d")

        # loop over points
        for j in range(npts):
            pt_cur = pts[j]
            wc_list = pt_dict[pt_cur]

            # loop over expansion functions
            for i in range(nexp):
                for (w, c) in wc_list:
                    result[c][i] += w * bfs[i, j]

        if self.deriv_dict:
            dpt_dict = self.deriv_dict

            # this makes things quicker since it uses dmats after
            # instantiation
            es_foo = polynomial_set.ONPolynomialSet(self.ref_el, ed)
            dpts = list(dpt_dict.keys())

            dbfs = es_foo.tabulate(dpts, self.max_deriv_order)

            ndpts = len(dpts)
            for j in range(ndpts):
                dpt_cur = dpts[j]
                wac_list = dpt_dict[dpt_cur]
                for i in range(nexp):
                    for (w, alpha, c) in wac_list:
                        result[c][i] += w * dbfs[tuple(alpha)][i, j]

        return result

    def tostr(self):
        return self.functional_type


class PointEvaluation(Functional):
    """Class representing point evaluation of scalar functions at a
    particular point x."""

    def __init__(self, ref_el, x):
        pt_dict = {tuple(x): [(1.0, tuple())]}
        super().__init__(ref_el, tuple(), pt_dict, {}, "PointEval")

    def __call__(self, fn):
        """Evaluate the functional on the function fn."""
        return fn(tuple(self.pt_dict.keys())[0])

    def tostr(self):
        x = list(map(str, list(self.pt_dict.keys())[0]))
        return "u(%s)" % (','.join(x),)


class ComponentPointEvaluation(Functional):
    """Class representing point evaluation of a particular component
    of a vector/tensor function at a particular point x."""

    def __init__(self, ref_el, comp, shp, x):
        if not isinstance(comp, tuple):
            comp = (comp,)
        if len(shp) != len(comp):
            raise ValueError("Component and shape are incompatible")
        if any(i < 0 or i >= n for i, n in zip(comp, shp)):
            raise ValueError("Illegal component")
        self.comp = comp
        pt_dict = {tuple(x): [(1.0, comp)]}
        super().__init__(ref_el, shp, pt_dict, {}, "ComponentPointEval")

    def tostr(self):
        x = list(map(str, list(self.pt_dict.keys())[0]))
        return "(u[%d](%s)" % (self.comp, ','.join(x))


class PointDerivative(Functional):
    """Class representing point partial differentiation of scalar
    functions at a particular point x."""

    def __init__(self, ref_el, x, alpha):
        dpt_dict = {x: [(1.0, tuple(alpha), tuple())]}
        self.alpha = tuple(alpha)
        self.order = sum(self.alpha)

        super().__init__(ref_el, tuple(), {}, dpt_dict, "PointDeriv")

    def __call__(self, fn):
        """Evaluate the functional on the function fn. Note that this depends
        on sympy being able to differentiate fn."""
        import sympy
        x, = self.deriv_dict

        X = tuple(sympy.Symbol(f"X[{i}]") for i in range(len(x)))

        dvars = tuple(d for d, a in zip(X, self.alpha)
                      for count in range(a))

        df = sympy.lambdify(X, sympy.diff(fn(X), *dvars))
        return df(*x)


class PointDirectionalDerivative(Functional):
    """Represents d/ds at a point."""
    def __init__(self, ref_el, s, pt, comp=(), shp=(), nm=None):
        sd = ref_el.get_spatial_dimension()
        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {pt: [(s[i], tuple(alphas[i]), comp) for i in range(sd)]}

        super().__init__(ref_el, shp, {}, dpt_dict, nm or "PointDirectionalDeriv")


class PointNormalDerivative(PointDirectionalDerivative):
    """Represents d/dn at a point on a facet."""
    def __init__(self, ref_el, facet_no, pt, comp=(), shp=()):
        n = ref_el.compute_normal(facet_no)
        super().__init__(ref_el, n, pt, comp=comp, shp=shp, nm="PointNormalDeriv")


class PointTangentialDerivative(PointDirectionalDerivative):
    """Represents d/dt at a point on an edge."""
    def __init__(self, ref_el, edge_no, pt, comp=(), shp=()):
        t = ref_el.compute_edge_tangent(edge_no)
        super().__init__(ref_el, t, pt, comp=comp, shp=shp, nm="PointTangentialDeriv")


class PointSecondDerivative(Functional):
    """Represents d/ds1 d/ds2 at a point."""
    def __init__(self, ref_el, s1, s2, pt, comp=(), shp=(), nm=None):
        sd = ref_el.get_spatial_dimension()
        tau = numpy.zeros((sd*(sd+1)//2,))

        alphas = []
        cur = 0
        for i in range(sd):
            for j in range(i, sd):
                alpha = [0] * sd
                alpha[i] += 1
                alpha[j] += 1
                alphas.append(tuple(alpha))
                tau[cur] = s1[i] * s2[j] + (i != j) * s2[i] * s1[j]
                cur += 1

        dpt_dict = {tuple(pt): [(tau[i], alphas[i], comp) for i in range(len(alphas))]}

        super().__init__(ref_el, shp, {}, dpt_dict, nm or "PointSecondDeriv")


class PointNormalSecondDerivative(PointSecondDerivative):
    """Represents d^/dn^2 at a point on a facet."""
    def __init__(self, ref_el, facet_no, pt, comp=(), shp=()):
        n = ref_el.compute_normal(facet_no)
        super().__init__(ref_el, n, n, pt, comp=comp, shp=shp, nm="PointNormalSecondDeriv")


class PointTangentialSecondDerivative(PointSecondDerivative):
    """Represents d^/dt^2 at a point on an edge."""
    def __init__(self, ref_el, edge_no, pt, comp=(), shp=()):
        t = ref_el.compute_edge_tangent(edge_no)
        super().__init__(ref_el, t, t, pt, comp=comp, shp=shp, nm="PointTangentialSecondDeriv")


class PointDivergence(Functional):
    """Class representing point divergence of vector
    functions at a particular point x."""

    def __init__(self, ref_el, x):
        sd = ref_el.get_spatial_dimension()
        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {x: [(1.0, alpha, (alpha.index(1),)) for alpha in alphas]}

        super().__init__(ref_el, (len(x),), {}, dpt_dict, "PointDiv")


class IntegralMoment(Functional):
    """Functional representing integral of the input against some tabulated function f.

    :arg ref_el: a :class:`Cell`.
    :arg Q: a :class:`QuadratureRule`.
    :arg f_at_qpts: an array tabulating the function f at the quadrature
         points.
    :arg comp: Optional argument indicating that only a particular
         component of the input function should be integrated against f
    :arg shp: Optional argument giving the value shape of input functions.
    """

    def __init__(self, ref_el, Q, f_at_qpts, comp=tuple(), shp=tuple()):
        self.Q = Q
        self.f_at_qpts = f_at_qpts
        self.comp = comp
        points = Q.get_points()
        weights = numpy.multiply(f_at_qpts, Q.get_weights())
        pt_dict = {tuple(pt): [(wt, comp)] for pt, wt in zip(points, weights)}
        super().__init__(ref_el, shp, pt_dict, {}, "IntegralMoment")

    def __call__(self, fn):
        """Evaluate the functional on the function fn."""
        pts = list(self.pt_dict.keys())
        wts = numpy.array([foo[0][0] for foo in list(self.pt_dict.values())])
        result = numpy.dot([fn(p) for p in pts], wts)

        if self.comp:
            result = result[self.comp]
        return result


class IntegralMomentOfDerivative(Functional):
    """Functional giving directional derivative integrated against some function on a facet."""

    def __init__(self, ref_el, s, Q, f_at_qpts, comp=(), shp=()):
        self.f_at_qpts = f_at_qpts
        self.Q = Q

        sd = ref_el.get_spatial_dimension()

        points = Q.get_points()
        weights = numpy.multiply(f_at_qpts, Q.get_weights())

        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {tuple(pt): [(wt*s[i], alphas[i], comp) for i in range(sd)]
                    for pt, wt in zip(points, weights)}

        super().__init__(ref_el, shp,
                         {}, dpt_dict, "IntegralMomentOfDerivative")


class IntegralMomentOfNormalDerivative(Functional):
    """Functional giving normal derivative integrated against some function on a facet."""

    def __init__(self, ref_el, facet_no, Q, f_at_qpts):
        n = ref_el.compute_normal(facet_no)
        self.n = n
        self.f_at_qpts = f_at_qpts
        self.Q = Q

        sd = ref_el.get_spatial_dimension()

        # map points onto facet
        transform = ref_el.get_entity_transform(sd-1, facet_no)
        points = transform(Q.get_points())
        self.dpts = points
        weights = numpy.multiply(f_at_qpts, Q.get_weights())

        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {tuple(pt): [(wt*n[i], alphas[i], tuple()) for i in range(sd)]
                    for pt, wt in zip(points, weights)}

        super().__init__(ref_el, tuple(),
                         {}, dpt_dict, "IntegralMomentOfNormalDerivative")


class FrobeniusIntegralMoment(IntegralMoment):

    def __init__(self, ref_el, Q, f_at_qpts, nm=None):
        # f_at_qpts is (some shape) x num_qpts
        shp = tuple(f_at_qpts.shape[:-1])
        if len(Q.pts) != f_at_qpts.shape[-1]:
            raise Exception("Mismatch in number of quadrature points and values")

        self.Q = Q
        self.comp = slice(None, None)
        self.f_at_qpts = f_at_qpts
        qpts, qwts = Q.get_points(), Q.get_weights()
        weights = numpy.transpose(numpy.multiply(f_at_qpts, qwts), (-1,) + tuple(range(len(shp))))
        alphas = list(index_iterator(shp))

        pt_dict = {tuple(pt): [(wt[alpha], alpha) for alpha in alphas] for pt, wt in zip(qpts, weights)}
        Functional.__init__(self, ref_el, shp, pt_dict, {}, nm or "FrobeniusIntegralMoment")


class IntegralLegendreDirectionalMoment(FrobeniusIntegralMoment):
    """Moment of v.s against a Legendre polynomial over an edge"""
    def __init__(self, cell, s, entity, mom_deg, quad_deg, nm=""):
        # mom_deg is degree of moment, quad_deg is the total degree of
        # polynomial you might need to integrate (or something like that)
        assert cell.get_spatial_dimension() == 2
        entity = (1, entity)

        Q = quadrature_schemes.create_quadrature(cell, quad_deg, entity=entity)
        x = cell.compute_barycentric_coordinates(Q.get_points(), entity=entity)

        f_at_qpts = jacobi.eval_jacobi(0, 0, mom_deg, x[:, 1] - x[:, 0])
        f_at_qpts /= Q.jacobian_determinant()

        f_at_qpts = numpy.multiply(s[..., None], f_at_qpts)
        super().__init__(cell, Q, f_at_qpts, nm=nm)


class IntegralLegendreNormalMoment(IntegralLegendreDirectionalMoment):
    """Moment of v.n against a Legendre polynomial over an edge"""
    def __init__(self, cell, entity, mom_deg, comp_deg):
        n = cell.compute_scaled_normal(entity)
        super().__init__(cell, n, entity, mom_deg, comp_deg,
                         "IntegralLegendreNormalMoment")


class IntegralLegendreTangentialMoment(IntegralLegendreDirectionalMoment):
    """Moment of v.t against a Legendre polynomial over an edge"""
    def __init__(self, cell, entity, mom_deg, comp_deg):
        t = cell.compute_edge_tangent(entity)
        super().__init__(cell, t, entity, mom_deg, comp_deg,
                         "IntegralLegendreTangentialMoment")


class IntegralLegendreBidirectionalMoment(IntegralLegendreDirectionalMoment):
    """Moment of dot(s1, dot(tau, s2)) against Legendre on entity, multiplied by the size of the reference facet"""
    def __init__(self, cell, s1, s2, entity, mom_deg, comp_deg, nm=""):
        s1s2T = numpy.outer(s1, s2)
        super().__init__(cell, s1s2T, entity, mom_deg, comp_deg, nm=nm)


class IntegralLegendreNormalNormalMoment(IntegralLegendreBidirectionalMoment):
    """Moment of dot(n, dot(tau, n)) against Legendre on entity."""
    def __init__(self, cell, entity, mom_deg, comp_deg):
        n = cell.compute_scaled_normal(entity)
        super().__init__(cell, n, n, entity, mom_deg, comp_deg,
                         "IntegralNormalNormalLegendreMoment")


class IntegralLegendreNormalTangentialMoment(IntegralLegendreBidirectionalMoment):
    """Moment of dot(n, dot(tau, t)) against Legendre on entity."""
    def __init__(self, cell, entity, mom_deg, comp_deg):
        n = cell.compute_scaled_normal(entity)
        t = cell.compute_edge_tangent(entity)
        super().__init__(cell, n, t, entity, mom_deg, comp_deg,
                         "IntegralNormalTangentialLegendreMoment")


class IntegralLegendreTangentialTangentialMoment(IntegralLegendreBidirectionalMoment):
    """Moment of dot(t, dot(tau, t)) against Legendre on entity."""
    def __init__(self, cell, entity, mom_deg, comp_deg):
        t = cell.compute_edge_tangent(entity)
        super().__init__(cell, t, t, entity, mom_deg, comp_deg,
                         "IntegralTangentialTangentialLegendreMoment")


class IntegralMomentOfDivergence(Functional):
    """Functional representing integral of the divergence of the input
    against some tabulated function f."""
    def __init__(self, ref_el, Q, f_at_qpts):
        self.f_at_qpts = f_at_qpts
        self.Q = Q

        sd = ref_el.get_spatial_dimension()

        points = Q.get_points()
        self.dpts = points
        weights = numpy.multiply(f_at_qpts, Q.get_weights())

        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {tuple(pt): [(wt, alphas[i], (i,)) for i in range(sd)]
                    for pt, wt in zip(points, weights)}

        super().__init__(ref_el, tuple(), {}, dpt_dict,
                         "IntegralMomentOfDivergence")


class IntegralMomentOfTensorDivergence(Functional):
    """Like IntegralMomentOfDivergence, but on symmetric tensors."""

    def __init__(self, ref_el, Q, f_at_qpts):
        self.f_at_qpts = f_at_qpts
        self.Q = Q
        points = Q.get_points()
        self.dpts = points
        sd = ref_el.get_spatial_dimension()
        shp = (sd, sd)

        assert len(f_at_qpts.shape) == 2
        assert f_at_qpts.shape[0] == sd
        assert f_at_qpts.shape[1] == len(points)
        weights = numpy.multiply(f_at_qpts, Q.get_weights()).T

        alphas = tuple(map(tuple, numpy.eye(sd, dtype=int)))
        dpt_dict = {tuple(pt): [(wt[i], alphas[j], (i, j)) for i, j in index_iterator(shp)]
                    for pt, wt in zip(points, weights)}

        super().__init__(ref_el, tuple(), {}, dpt_dict, "IntegralMomentOfDivergence")


class PointNormalEvaluation(Functional):
    """Implements the evaluation of the normal component of a vector at a
    point on a facet of codimension 1."""

    def __init__(self, ref_el, facet_no, pt):
        n = ref_el.compute_normal(facet_no)
        self.n = n
        shp = n.shape
        pt_dict = {pt: [(n[i], (i,)) for i in range(shp[0])]}
        super().__init__(ref_el, shp, pt_dict, {}, "PointNormalEval")


class PointEdgeTangentEvaluation(Functional):
    """Implements the evaluation of the tangential component of a
    vector at a point on a facet of dimension 1."""

    def __init__(self, ref_el, edge_no, pt):
        t = ref_el.compute_edge_tangent(edge_no)
        self.t = t
        shp = t.shape
        pt_dict = {pt: [(t[i], (i,)) for i in range(shp[0])]}
        super().__init__(ref_el, shp, pt_dict, {}, "PointEdgeTangent")

    def tostr(self):
        x = list(map(str, list(self.pt_dict.keys())[0]))
        return "(u.t)(%s)" % (','.join(x),)


class IntegralMomentOfEdgeTangentEvaluation(Functional):
    r"""
    \int_e v\cdot t p ds

    p \in Polynomials

    :arg ref_el: reference element for which e is a dim-1 entity
    :arg Q: quadrature rule on the face
    :arg P_at_qpts: polynomials evaluated at quad points
    :arg edge: which edge.
    """
    def __init__(self, ref_el, Q, P_at_qpts, edge):
        t = ref_el.compute_edge_tangent(edge)
        sd = ref_el.get_spatial_dimension()
        transform = ref_el.get_entity_transform(1, edge)
        points = transform(Q.get_points())
        weights = numpy.multiply(P_at_qpts, Q.get_weights())
        pt_dict = {tuple(pt): [(wt*t[i], (i,)) for i in range(sd)]
                   for pt, wt in zip(points, weights)}
        super().__init__(ref_el, (sd, ), pt_dict, {},
                         "IntegralMomentOfEdgeTangentEvaluation")


class PointFaceTangentEvaluation(Functional):
    """Implements the evaluation of a tangential component of a
    vector at a point on a facet of codimension 1."""

    def __init__(self, ref_el, face_no, tno, pt):
        t = ref_el.compute_face_tangents(face_no)[tno]
        self.t = t
        self.tno = tno
        sd = ref_el.get_spatial_dimension()
        pt_dict = {pt: [(t[i], (i,)) for i in range(sd)]}
        shp = (sd,)
        Functional.__init__(self, ref_el, shp, pt_dict, {}, "PointFaceTangent")

    def tostr(self):
        x = list(map(str, list(self.pt_dict.keys())[0]))
        return "(u.t%d)(%s)" % (self.tno, ','.join(x),)


class IntegralMomentOfFaceTangentEvaluation(Functional):
    r"""
    \int_F v \times n \cdot p ds

    p \in Polynomials

    :arg ref_el: reference element for which F is a codim-1 entity
    :arg Q: quadrature rule on the face
    :arg P_at_qpts: polynomials evaluated at quad points
    :arg facet: which facet.
    """
    def __init__(self, ref_el, Q, P_at_qpts, facet):
        P_at_qpts = [[P_at_qpts[0][i], P_at_qpts[1][i], P_at_qpts[2][i]]
                     for i in range(P_at_qpts.shape[1])]
        n = ref_el.compute_scaled_normal(facet)
        sd = ref_el.get_spatial_dimension()
        transform = ref_el.get_entity_transform(sd-1, facet)
        pts = tuple(map(tuple, transform(Q.get_points())))
        weights = Q.get_weights()
        pt_dict = {}
        for pt, wgt, phi in zip(pts, weights, P_at_qpts):
            phixn = [phi[1]*n[2] - phi[2]*n[1],
                     phi[2]*n[0] - phi[0]*n[2],
                     phi[0]*n[1] - phi[1]*n[0]]
            pt_dict[pt] = [(wgt*(-n[2]*phixn[1]+n[1]*phixn[2]), (0, )),
                           (wgt*(n[2]*phixn[0]-n[0]*phixn[2]), (1, )),
                           (wgt*(-n[1]*phixn[0]+n[0]*phixn[1]), (2, ))]
        super().__init__(ref_el, (sd, ), pt_dict, {},
                         "IntegralMomentOfFaceTangentEvaluation")


class PointScaledNormalEvaluation(Functional):
    """Implements the evaluation of the normal component of a vector at a
    point on a facet of codimension 1, where the normal is scaled by
    the volume of that facet."""

    def __init__(self, ref_el, facet_no, pt):
        n = ref_el.compute_scaled_normal(facet_no)
        sd = ref_el.get_spatial_dimension()
        shp = (sd,)

        pt_dict = {pt: [(n[i], (i,)) for i in range(sd)]}
        super().__init__(ref_el, shp, pt_dict, {}, "PointScaledNormalEval")

    def tostr(self):
        x = list(map(str, list(self.pt_dict.keys())[0]))
        return "(u.n)(%s)" % (','.join(x),)


class IntegralMomentOfScaledNormalEvaluation(Functional):
    r"""
    \int_F v\cdot n p ds

    p \in Polynomials

    :arg ref_el: reference element for which F is a codim-1 entity
    :arg Q: quadrature rule on the face
    :arg P_at_qpts: polynomials evaluated at quad points
    :arg facet: which facet.
    """
    def __init__(self, ref_el, Q, P_at_qpts, facet):
        n = ref_el.compute_scaled_normal(facet)
        sd = ref_el.get_spatial_dimension()
        transform = ref_el.get_entity_transform(sd - 1, facet)
        pts = transform(Q.get_points())
        weights = Q.get_weights() * P_at_qpts
        pt_dict = {tuple(pt): [(wt*n[i], (i, )) for i in range(sd)]
                   for pt, wt in zip(pts, weights)}
        super().__init__(ref_el, (sd, ), pt_dict, {}, "IntegralMomentOfScaledNormalEvaluation")


class PointwiseInnerProductEvaluation(Functional):
    """
    This is a functional on symmetric 2-tensor fields. Let u be such a
    field, p be a point, and v,w be vectors. This implements the evaluation
    v^T u(p) w.

    Clearly v^iu_{ij}w^j = u_{ij}v^iw^j. Thus the value can be computed
    from the Frobenius inner product of u with wv^T. This gives the
    correct weights.
    """

    def __init__(self, ref_el, v, w, pt):
        wvT = numpy.outer(w, v)
        shp = wvT.shape

        pt_dict = {tuple(pt): [(wvT[idx], idx) for idx in index_iterator(shp)]}

        super().__init__(ref_el, shp, pt_dict, {}, "PointwiseInnerProductEval")


class TensorBidirectionalIntegralMoment(FrobeniusIntegralMoment):
    r"""
    This is a functional on symmetric 2-tensor fields. Let u be such a
    field, f a function tabulated at points, and v,w be vectors. This implements the evaluation
    \int v^T u(x) w f(x).
    Clearly v^iu_{ij}w^j = u_{ij}v^iw^j. Thus the value can be computed
    from the Frobenius inner product of u with vw^T. This gives the
    correct weights.
    """

    def __init__(self, ref_el, v, w, Q, f_at_qpts):
        vwT = numpy.outer(v, w)
        F_at_qpts = numpy.multiply(vwT[..., None], f_at_qpts)
        super().__init__(ref_el, Q, F_at_qpts, "TensorBidirectionalMomentInnerProductEvaluation")


class IntegralMomentOfNormalEvaluation(Functional):
    r"""
    \int_F v\cdot n p ds
    p \in Polynomials
    :arg ref_el: reference element for which F is a codim-1 entity
    :arg Q: quadrature rule on the face
    :arg P_at_qpts: polynomials evaluated at quad points
    :arg facet: which facet.
    """
    def __init__(self, ref_el, Q, P_at_qpts, facet):
        # scaling on the normal is ok because edge length then weights
        # the reference element quadrature appropriately
        n = ref_el.compute_scaled_normal(facet)
        sd = ref_el.get_spatial_dimension()
        transform = ref_el.get_entity_transform(sd - 1, facet)
        pts = transform(Q.get_points())
        weights = numpy.multiply(P_at_qpts, Q.get_weights())
        pt_dict = {tuple(pt): [(wt*n[i], (i, )) for i in range(sd)]
                   for pt, wt in zip(pts, weights)}
        super().__init__(ref_el, (sd, ), pt_dict, {}, "IntegralMomentOfNormalEvaluation")


class IntegralMomentOfTangentialEvaluation(Functional):
    r"""
    \int_F v\cdot n p ds
    p \in Polynomials
    :arg ref_el: reference element for which F is a codim-1 entity
    :arg Q: quadrature rule on the face
    :arg P_at_qpts: polynomials evaluated at quad points
    :arg facet: which facet.
    """
    def __init__(self, ref_el, Q, P_at_qpts, facet):
        # scaling on the tangent is ok because edge length then weights
        # the reference element quadrature appropriately
        sd = ref_el.get_spatial_dimension()
        assert sd == 2
        t = ref_el.compute_edge_tangent(facet)
        transform = ref_el.get_entity_transform(sd - 1, facet)
        points = transform(Q.get_points())
        weights = numpy.multiply(P_at_qpts, Q.get_weights())
        pt_dict = {tuple(pt): [(wt*t[i], (i, )) for i in range(sd)]
                   for pt, wt in zip(points, weights)}
        super().__init__(ref_el, (sd, ), pt_dict, {}, "IntegralMomentOfScaledTangentialEvaluation")



"""Generic abstract node class and utility functions for creating
expression DAG languages."""

import collections
import gem


class Node(object):
    """Abstract node class.

    Nodes are not meant to be modified.

    A node can reference other nodes; they are called children. A node
    might contain data, or reference other objects which are not
    themselves nodes; they are not called children.

    Both the children (if any) and non-child data (if any) are
    required to create a node, or determine the equality of two
    nodes. For reconstruction, however, only the new children are
    necessary.
    """

    __slots__ = ('hash_value',)

    # Non-child data as the first arguments of the constructor.
    # To be (potentially) overridden by derived node classes.
    __front__ = ()

    # Non-child data as the last arguments of the constructor.
    # To be (potentially) overridden by derived node classes.
    __back__ = ()

    def _cons_args(self, children):
        """Constructs an argument list for the constructor with
        non-child data from 'self' and children from 'children'.

        Internally used utility function.
        """
        front_args = [getattr(self, name) for name in self.__front__]
        back_args = [getattr(self, name) for name in self.__back__]

        return tuple(front_args) + tuple(children) + tuple(back_args)

    def __reduce__(self):
        # Gold version:
        return type(self), self._cons_args(self.children)

    def reconstruct(self, *args):
        """Reconstructs the node with new children from
        'args'. Non-child data are copied from 'self'.

        Returns a new object.
        """
        return type(self)(*self._cons_args(args))

    def __repr__(self):
        cons_args = self._cons_args(self.children)
        return "%s(%s)" % (type(self).__name__, ", ".join(map(repr, cons_args)))

    def __eq__(self, other):
        """Provides equality testing with quick positive and negative
        paths based on :func:`id` and :meth:`__hash__`.
        """
        if self is other:
            return True
        elif hash(self) != hash(other):
            return False
        else:
            return self.is_equal(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        """Provides caching for hash values."""
        try:
            return self.hash_value
        except AttributeError:
            self.hash_value = self.get_hash()
            return self.hash_value

    def is_equal(self, other):
        """Equality predicate.

        This is the method to potentially override in derived classes,
        not :meth:`__eq__` or :meth:`__ne__`.
        """
        if type(self) is not type(other):
            return False
        self_consargs = self._cons_args(self.children)
        other_consargs = other._cons_args(other.children)
        return self_consargs == other_consargs

    def get_hash(self):
        """Hash function.

        This is the method to potentially override in derived classes,
        not :meth:`__hash__`.
        """
        return hash((type(self),) + self._cons_args(self.children))


def _make_traversal_children(node):
    if isinstance(node, (gem.Indexed, gem.FlexiblyIndexed)):
        # Include child nodes hidden in index expressions.
        return node.children + node.indirect_children
    else:
        return node.children


def pre_traversal(expression_dags):
    """Pre-order traversal of the nodes of expression DAGs.

    Notes
    -----
    This function also walks through nodes in index expressions
    (e.g., `VariableIndex`s); see ``_make_traversal_children()``.

    """
    seen = set()
    lifo = []
    # Some roots might be same, but they must be visited only once.
    # Keep the original ordering of roots, for deterministic code
    # generation.
    for root in expression_dags:
        if root not in seen:
            seen.add(root)
            lifo.append(root)

    while lifo:
        node = lifo.pop()
        yield node
        children = _make_traversal_children(node)
        for child in reversed(children):
            if child not in seen:
                seen.add(child)
                lifo.append(child)


def post_traversal(expression_dags):
    """Post-order traversal of the nodes of expression DAGs.

    Notes
    -----
    This function also walks through nodes in index expressions
    (e.g., `VariableIndex`s); see ``_make_traversal_children()``.


    """
    seen = set()
    lifo = []
    # Some roots might be same, but they must be visited only once.
    # Keep the original ordering of roots, for deterministic code
    # generation.
    for root in expression_dags:
        if root not in seen:
            seen.add(root)
            lifo.append((root, list(_make_traversal_children(root))))

    while lifo:
        node, deps = lifo[-1]
        for i, dep in enumerate(deps):
            if dep is not None and dep not in seen:
                lifo.append((dep, list(_make_traversal_children(dep))))
                deps[i] = None
                break
        else:
            yield node
            seen.add(node)
            lifo.pop()


# Default to the more efficient pre-order traversal
traversal = pre_traversal


def collect_refcount(expression_dags):
    """Collects reference counts for a multi-root expression DAG.

    Notes
    -----
    This function also collects reference counts of nodes
    in index expressions (e.g., `VariableIndex`s); see
    ``_make_traversal_children()``.

    """
    result = collections.Counter(expression_dags)
    for node in traversal(expression_dags):
        result.update(_make_traversal_children(node))
    return result


def noop_recursive(function):
    """No-op wrapper for functions with overridable recursive calls.

    :arg function: a function with parameters (value, rec), where
                   ``rec`` is expected to be a function used for
                   recursive calls.
    :returns: a function with working recursion and nothing fancy
    """
    def recursive(node):
        return function(node, recursive)
    return recursive


def noop_recursive_arg(function):
    """No-op wrapper for functions with overridable recursive calls
    and an argument.

    :arg function: a function with parameters (value, rec, arg), where
                   ``rec`` is expected to be a function used for
                   recursive calls.
    :returns: a function with working recursion and nothing fancy
    """
    def recursive(node, arg):
        return function(node, recursive, arg)
    return recursive


class Memoizer(object):
    """Caching wrapper for functions with overridable recursive calls.
    The lifetime of the cache is the lifetime of the object instance.

    :arg function: a function with parameters (value, rec), where
                   ``rec`` is expected to be a function used for
                   recursive calls.
    :returns: a function with working recursion and caching
    """
    def __init__(self, function):
        self.cache = {}
        self.function = function

    def __call__(self, node):
        try:
            return self.cache[node]
        except KeyError:
            result = self.function(node, self)
            self.cache[node] = result
            return result


class MemoizerArg(object):
    """Caching wrapper for functions with overridable recursive calls
    and an argument.  The lifetime of the cache is the lifetime of the
    object instance.

    :arg function: a function with parameters (value, rec, arg), where
                   ``rec`` is expected to be a function used for
                   recursive calls.
    :returns: a function with working recursion and caching
    """
    def __init__(self, function):
        self.cache = {}
        self.function = function

    def __call__(self, node, arg):
        cache_key = (node, arg)
        try:
            return self.cache[cache_key]
        except KeyError:
            result = self.function(node, self, arg)
            self.cache[cache_key] = result
            return result


def reuse_if_untouched(node, self):
    """Reuse if untouched recipe"""
    new_children = list(map(self, node.children))
    if all(nc == c for nc, c in zip(new_children, node.children)):
        return node
    else:
        return node.reconstruct(*new_children)


def reuse_if_untouched_arg(node, self, arg):
    """Reuse if touched recipe propagating an extra argument"""
    new_children = [self(child, arg) for child in node.children]
    if all(nc == c for nc, c in zip(new_children, node.children)):
        return node
    else:
        return node.reconstruct(*new_children)
